/**
 * DBEC-GP-OMP-CUDA-MPI programs are developed by:
 *
 * Vladimir Loncar, Antun Balaz
 * (Scientific Computing Laboratory, Institute of Physics Belgrade, Serbia)
 *
 * Srdjan Skrbic
 * (Department of Mathematics and Informatics, Faculty of Sciences, University of Novi Sad, Serbia)
 *
 * Paulsamy Muruganandam
 * (Bharathidasan University, Tamil Nadu, India)
 *
 * Luis E. Young-S, Sadhan K. Adhikari
 * (UNESP - Sao Paulo State University, Brazil)
 *
 *
 * Public use and modification of these codes are allowed provided that the
 * following papers are cited:
 * [1] V. Loncar et al., Comput. Phys. Commun. 209 (2016) 190.      
 * [2] V. Loncar et al., Comput. Phys. Commun. 200 (2016) 406.      
 * [3] R. Kishor Kumar et al., Comput. Phys. Commun. 195 (2015) 117.
 *
 * The authors would be grateful for all information and/or comments
 * regarding the use of the programs.
 */

#include "imag1dZ-th.h"

int main(int argc, char **argv) {
   FILE *out;
   FILE *file;
   char filename[MAX_FILENAME_SIZE];
   int nthreads;
   long cnti;
   double norm, mu, en;
   double *rms;
   double *psi;
   double *psidd2;
   double *cbeta;
   double *dpsiz;
   double *tmpzi, *tmpzj, *tmpzk, *tmpzl, *tmpzm, *tmpzn;
   fftw_complex *psidd2fft;
   double psi2;

   pi = 3.14159265358979;

   struct timeval start, stop, iter_start, iter_stop;
   double wall_time, init_time, iter_time;
   iter_time = 0.;
   gettimeofday(&start, NULL);

   if((argc != 3) || (strcmp(*(argv + 1), "-i") != 0)) {
      fprintf(stderr, "Usage: %s -i <input parameter file> \n", *argv);
      exit(EXIT_FAILURE);
   }

   if(! cfg_init(argv[2])) {
      fprintf(stderr, "Wrong input parameter file.\n");
      exit(EXIT_FAILURE);
   }

   readpar();

   #pragma omp parallel
   #pragma omp master
   nthreads = omp_get_num_threads();

   rms = alloc_double_vector(RMS_ARRAY_SIZE);

   z = alloc_double_vector(Nz);

   z2 = alloc_double_vector(Nz);

   pot = alloc_double_vector(Nz);
   potdd = alloc_double_vector(Nz);
   psi = alloc_double_vector(Nz);
   psidd2 = alloc_double_vector((Nz / 2 + 1) * 2);
   psidd2fft = (fftw_complex *) psidd2;

   dpsiz = alloc_double_vector(Nz);

   calphaz = alloc_double_vector(Nz - 1);
   cbeta =  alloc_double_vector(Nz - 1);
   cgammaz = alloc_double_vector(Nz - 1);

   tmpzi = alloc_double_vector(Nz);
   tmpzj = alloc_double_vector(Nz);
   tmpzk = alloc_double_vector(Nz);
   tmpzl = alloc_double_vector(Nz);
   tmpzm = alloc_double_vector(Nz);
   tmpzn = alloc_double_vector(Nz);

   if(output != NULL) {
      sprintf(filename, "%s.txt", output);
      out = fopen(filename, "w");
   }
   else out = stdout;

   if(opt == 2) par = 2.;
   else par = 1.;

   drho2 = drho * drho;
   g_1d = g0 / ((2. * pi * drho2));
   gd_1d = gd0 / ((2. * pi* drho2));

   Nz2 = Nz / 2;
   dz2 = dz * dz;

   fprintf(out, " Imaginary time propagation 1dZ,   OPTION = %d, OMP_NUM_THREADS = %d\n\n", opt, nthreads);
   fprintf(out, "  Number of Atoms N = %li, Unit of length AHO = %.8f m\n", Na,aho);
   fprintf(out, "  Scattering length a = %.2f*a0, Dipolar ADD = %.2f*a0\n", as, add);
   fprintf(out, "  Nonlinearity G_3D = %.4f, Strength of DDI GD_3D = %.5f\n", g0, gd0);
   fprintf(out, "  Nonlinearity G_1D = %.4f, Strength of DDI GD_1D = %.5f\n", g_1d, gd_1d);
   fprintf(out, "  Parameter of trap: LAMBDA = %.2f\n", vlambda);
   fprintf(out, "  Radial Gaussian Width = %.2f\n\n", drho);
   fprintf(out, " # Space Stp:  N = %li\n", Nz);
   fprintf(out, "  Space Step: DZ = %.6f\n", dz);
   fprintf(out, " # Time Stp : NSTP = %li, NPAS = %li, NRUN = %li\n", Nstp, Npas, Nrun);
   fprintf(out, "   Time Step:   DT = %.6f\n\n",  dt);
   fprintf(out, "                  --------------------------------------------------------\n");
   fprintf(out, "                    Norm      Chem        Ener/N      <z>      |Psi(0)|^2\n");
   fprintf(out, "                  --------------------------------------------------------\n");
   fflush(out);

   g = 0.;
   gd = 0.;

   fftw_init_threads();
   fftw_plan_with_nthreads(nthreads);
   initpsi(psi);
   initpot();
   gencoef();
   initpotdd(tmpzi, tmpzj, tmpzk, tmpzl, tmpzm, tmpzn);

   plan_forward = fftw_plan_dft_r2c_1d(Nz, psidd2, psidd2fft, FFT_FLAG);
   plan_backward = fftw_plan_dft_c2r_1d(Nz, psidd2fft, psidd2, FFT_FLAG);


   calcnorm(&norm, psi, tmpzi);
   calcmuen(&mu, &en, psi, dpsiz, psidd2, tmpzi, tmpzj);
   calcrms(rms, psi, tmpzi);
   psi2 = psi[Nz2] * psi[Nz2];
   fprintf(out, "Initial : %15.4f %11.5f %11.5f %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
   fflush(out);
   if(initout != NULL) {
      sprintf(filename, "%s.txt", initout);
      file = fopen(filename, "w");
      outpsi2z(psi, file);
      fclose(file);
   }

   if(Nstp != 0) {
      double g_stp = par * g_1d / (double) Nstp;
      double gd_stp = par * gd0 / (double) Nstp;
      g = 0.;
      gd = 0.;
      gettimeofday(&iter_start, NULL);
      for(cnti = 0; cnti < Nstp; cnti ++) {
         g += g_stp;
         gd += gd_stp;
         calcnu(psi, psidd2);
         calcluz(psi, cbeta);
         calcnorm(&norm, psi, tmpzi);
      }
      gettimeofday(&iter_stop, NULL);
      iter_time += (double) (((iter_stop.tv_sec - iter_start.tv_sec) * 1000 + (iter_stop.tv_usec - iter_start.tv_usec)/1000.0) + 0.5);

      calcmuen(&mu, &en, psi, dpsiz, psidd2, tmpzi, tmpzj);
      calcrms(rms, psi, tmpzi);
      psi2 = psi[Nz2] * psi[Nz2];
      fprintf(out, "After NSTP iter.:%8.4f %11.5f %11.5f %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
      fflush(out);
      if(Nstpout != NULL) {
         sprintf(filename, "%s.txt", Nstpout);
         file = fopen(filename, "w");
         outpsi2z(psi, file);
         fclose(file);
      }
   }
   else {
      g = par * g_1d;
      gd = par * gd0;
   }

   gettimeofday(&iter_start, NULL);
   for(cnti = 0; cnti < Npas; cnti ++) {
      calcnu(psi, psidd2);
      calcluz(psi, cbeta);
      calcnorm(&norm, psi, tmpzi);
   }
   gettimeofday(&iter_stop, NULL);
   iter_time += (double) (((iter_stop.tv_sec - iter_start.tv_sec) * 1000 + (iter_stop.tv_usec - iter_start.tv_usec)/1000.0) + 0.5);

   calcmuen(&mu, &en, psi, dpsiz, psidd2, tmpzi, tmpzj);
   calcrms(rms, psi, tmpzi);
   psi2 = psi[Nz2] * psi[Nz2];
   fprintf(out, "After NPAS iter.:%8.4f %11.5f %11.5f %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
   fflush(out);
   if(Npasout != NULL) {
      sprintf(filename, "%s.txt", Npasout);
      file = fopen(filename, "w");
      outpsi2z(psi, file);
      fclose(file);
   }

   gettimeofday(&iter_start, NULL);
   for(cnti = 0; cnti < Nrun; cnti ++) {
      calcnu(psi, psidd2);
      calcluz(psi, cbeta);
      calcnorm(&norm, psi, tmpzi);
   }
   gettimeofday(&iter_stop, NULL);
   iter_time += (double) (((iter_stop.tv_sec - iter_start.tv_sec) * 1000 + (iter_stop.tv_usec - iter_start.tv_usec)/1000.0) + 0.5);

   calcmuen(&mu, &en, psi, dpsiz, psidd2, tmpzi, tmpzj);
   calcrms(rms, psi, tmpzi);
   psi2 = psi[Nz2] * psi[Nz2];
   fprintf(out, "After NRUN iter.:%8.4f %11.5f %11.5f %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
   fflush(out);
   if(Nrunout != NULL) {
      sprintf(filename, "%s.txt", Nrunout);
      file = fopen(filename, "w");
      outpsi2z(psi, file);
      fclose(file);
   }
   fprintf(out, "                  --------------------------------------------------------\n\n");

   free_double_vector(rms);

   free_double_vector(z);

   free_double_vector(z2);

   free_double_vector(pot);
   free_double_vector(potdd);
   free_double_vector(psi);
   free_double_vector(psidd2);

   free_double_vector(dpsiz);

   free_double_vector(calphaz);
   free_double_vector(cbeta);
   free_double_vector(cgammaz);

   free_double_vector(tmpzi);
   free_double_vector(tmpzj);
   free_double_vector(tmpzk);
   free_double_vector(tmpzl);
   free_double_vector(tmpzm);
   free_double_vector(tmpzn);

   fftw_destroy_plan(plan_forward);
   fftw_destroy_plan(plan_backward);

   gettimeofday(&stop, NULL);
   wall_time = (double) (((stop.tv_sec - start.tv_sec) * 1000 + (stop.tv_usec - start.tv_usec)/1000.0) + 0.5);
   init_time = wall_time - iter_time;
   fprintf(out, " Initialization/allocation wall-clock time: %.3f seconds\n", init_time / 1000.);
   fprintf(out, " Calculation (iterations) wall-clock time: %.3f seconds\n", iter_time / 1000.);

   if(output != NULL) fclose(out);

   return(EXIT_SUCCESS);
}

/**
 *    Reading input parameters from the configuration file.
 */
void readpar(void) {
   char *cfg_tmp;

   if((cfg_tmp = cfg_read("OPTION")) == NULL) {
      fprintf(stderr, "OPTION is not defined in the configuration file\n");
      exit(EXIT_FAILURE);
   }
   opt = atol(cfg_tmp);

   if((cfg_tmp = cfg_read("NATOMS")) == NULL) {
      fprintf(stderr, "NATOMS is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Na = atol(cfg_tmp);

   if((cfg_tmp = cfg_read("G0")) == NULL) {
      if((cfg_tmp = cfg_read("AHO")) == NULL) {
         fprintf(stderr, "AHO is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      aho = atof(cfg_tmp);

      if((cfg_tmp = cfg_read("AS")) == NULL) {
         fprintf(stderr, "AS is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      as = atof(cfg_tmp);

      g0 = 4. * pi * as * Na * BOHR_RADIUS / aho;
   } else {
      g0 = atof(cfg_tmp);
   }

   if((cfg_tmp = cfg_read("GDD0")) == NULL) {
      if((cfg_tmp = cfg_read("AHO")) == NULL) {
         fprintf(stderr, "AHO is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      aho = atof(cfg_tmp);

      if((cfg_tmp = cfg_read("ADD")) == NULL) {
         fprintf(stderr, "ADD is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      add = atof(cfg_tmp);

      gd0 = 3. * add * Na * BOHR_RADIUS / aho;
   } else {
      gd0 = atof(cfg_tmp);
   }

   if((cfg_tmp = cfg_read("NZ")) == NULL) {
      fprintf(stderr, "NZ is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Nz = atol(cfg_tmp);

   if((cfg_tmp = cfg_read("DZ")) == NULL) {
      fprintf(stderr, "DZ is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   dz = atof(cfg_tmp);

   if((cfg_tmp = cfg_read("DT")) == NULL) {
      fprintf(stderr, "DT is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   dt = atof(cfg_tmp);

   if((cfg_tmp = cfg_read("LAMBDA")) == NULL) {
      fprintf(stderr, "LAMBDA is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   vlambda = atof(cfg_tmp);

   if((cfg_tmp = cfg_read("DRHO")) == NULL) {
      fprintf(stderr, "DRHO is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   drho = atof(cfg_tmp);

   if((cfg_tmp = cfg_read("NSTP")) == NULL) {
      fprintf(stderr, "NSTP is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Nstp = atol(cfg_tmp);

   if((cfg_tmp = cfg_read("NPAS")) == NULL) {
      fprintf(stderr, "NPAS is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Npas = atol(cfg_tmp);

   if((cfg_tmp = cfg_read("NRUN")) == NULL) {
      fprintf(stderr, "NRUN is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Nrun = atol(cfg_tmp);

   output = cfg_read("OUTPUT");
   initout = cfg_read("INITOUT");
   Nstpout = cfg_read("NSTPOUT");
   Npasout = cfg_read("NPASOUT");
   Nrunout = cfg_read("NRUNOUT");

   if((initout != NULL) || (Nstpout != NULL) || (Npasout != NULL) || (Nrunout != NULL)) {
      if((cfg_tmp = cfg_read("OUTSTPZ")) == NULL) {
         fprintf(stderr, "OUTSTPZ is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      outstpz = atol(cfg_tmp);
   }

   return;
}

/**
 *    Initialization of the space mesh and the initial wave function.
 *    psi - array with the wave function values
 */
void initpsi(double *psi) {
   long cnti;
   double cpsi;
   double tmp;

   cpsi = sqrt(sqrt(pi / vlambda));

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nz; cnti ++) {
      z[cnti] = (cnti - Nz2) * dz;
      z2[cnti] = z[cnti] * z[cnti];
   }

   #pragma omp parallel for private(cnti, tmp)
   for(cnti = 0; cnti < Nz; cnti ++) {
      tmp = exp(- 0.5 * vlambda * z2[cnti]);
      psi[cnti] = tmp / cpsi;
   }

   return;
}

/**
 *    Initialization of the potential.
 */
void initpot() {
   long cnti;
   double vlambda2;

   vlambda2 = vlambda * vlambda;

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nz; cnti ++) {
      pot[cnti] = 0.5 * par * (vlambda2 * z2[cnti]);
   }

   return;
}

/**
 *    Crank-Nicolson scheme coefficients generation.
 */
void gencoef(void) {
   long cnti;

   Az0 = 1. + dt / dz2 / (3. - par);

   Az0r = 1. - dt / dz2 / (3. - par);

   Az = - 0.5 * dt / dz2 / (3. - par);

   calphaz[Nz - 2] = 0.;
   cgammaz[Nz - 2] = - 1. / Az0;
   for (cnti = Nz - 2; cnti > 0; cnti --) {
      calphaz[cnti - 1] = Az * cgammaz[cnti];
      cgammaz[cnti - 1] = - 1. / (Az0 + Az * calphaz[cnti - 1]);
   }

   return;
}

/**
 *    Initialization of the dipolar potential.
 *    kz  - array with the space mesh values in the z-direction in the K-space
 *    kz2 - array with the squared space mesh values in the z-direction in the
 *          K-space
 */
void initpotdd(double *kz, double *kz2, double *zg, double *dzg, double *kr, double *dkr) {
   long cnti, cntj;
   double dkz;
   double kr2;

   gauleg(-1., 1., zg, dzg, Nz);

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nz; cnti ++) {
      kr[cnti] = (1. + zg[cnti]) / (1. - zg[cnti]);
      dkr[cnti] = 2. * dzg[cnti] / ((1. - zg[cnti]) * (1. - zg[cnti]));
   }

   dkz = 2. * pi / (Nz * dz);

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nz2; cnti ++) kz[cnti] = cnti * dkz;

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nz2; cnti ++) kz[cnti + Nz2] = (cnti - Nz2) * dkz;

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nz; cnti ++) kz2[cnti] = kz[cnti] * kz[cnti];

   #pragma omp parallel for private(cnti, cntj, kr2)
   for(cnti = 0; cnti < Nz; cnti ++) {
      potdd[cnti] = 0.;
      for(cntj = 0; cntj < Nz; cntj ++) {
         kr2 = kr[cntj] * kr[cntj];
         potdd[cnti] += (3. * kz2[cnti] / (kr2 + kz2[cnti]) - 1.) * kr[cntj] * dkr[cntj] * exp(- kr2 * drho2 / 2.);
      }
      potdd[cnti] *= 2. / 3.;
   }

   return;
}

/**
 *    Calculation of the wave function norm and normalization.
 *    norm - wave function norm
 *    psi  - array with the wave function values
 *    tmpz - temporary array
 */
void calcnorm(double *norm, double *psi, double *tmpz) {
   long cnti;
   double tmp;

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nz; cnti ++) {
      tmpz[cnti] = psi[cnti] * psi[cnti];
   }

   *norm = sqrt(simpint_th(dz, tmpz, Nz));
   tmp = 1. / *norm;

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nz; cnti ++) {
      psi[cnti] *= tmp;
   }

   return;
}

/**
 *    Calculation of the chemical potential and energy.
 *    mu     - chemical potential
 *    en     - energy
 *    psi    - array with the wave function values
 *    dpsiz  - temporary array
 *    psidd2 - array with the squared wave function values
 *    tmpzi  - temporary array
 *    tmpzj  - temporary array
 */
void calcmuen(double *mu, double *en, double *psi, double *dpsiz, double *psidd2, double *tmpzi, double *tmpzj) {
   long cnti;
   double psi2, psi2lin, psidd2lin, dpsi2;

   diff_th(dz, psi, dpsiz, Nz);

   calcpsidd2(psi, psidd2);

   #pragma omp parallel for private(cnti, psi2, psi2lin, psidd2lin, dpsi2)
   for(cnti = 0; cnti < Nz; cnti ++) {
      psi2 = psi[cnti] * psi[cnti];
      psi2lin = psi2 * g;
      psidd2lin = psidd2[cnti] * gd;
      dpsi2 = dpsiz[cnti] * dpsiz[cnti] / (3. - par);
      tmpzi[cnti] = (pot[cnti] + psi2lin + psidd2lin) * psi2 + dpsi2;
      tmpzj[cnti] = (pot[cnti] + 0.5 * psi2lin + 0.5 * psidd2lin) * psi2 + dpsi2;
   }

   *mu = simpint_th(dz, tmpzi, Nz);
   *en = simpint_th(dz, tmpzj, Nz);

   return;
}

/**
 *    Calculation of squared wave function values for dipole-dipole
 *    interaction.
 *    psi    - array with the wave function values
 *    psidd2 - array with the squared wave function values
 */
void calcpsidd2(double *psi, double *psidd2) {
   long cnti;
   fftw_complex *psidd2fft = (fftw_complex *) psidd2;

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nz; cnti ++) {
      psidd2[cnti] = psi[cnti] * psi[cnti];
   }

   fftw_execute_dft_r2c(plan_forward, psidd2, psidd2fft);

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nz2 + 1; cnti ++) {
      psidd2fft[cnti][0] *= potdd[cnti];
      psidd2fft[cnti][1] *= potdd[cnti];
   }

   fftw_execute_dft_c2r(plan_backward, psidd2fft, psidd2);

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nz; cnti ++) {
      psidd2[cnti] /= Nz;
   }

   psidd2[Nz - 1] = psidd2[0];

   return;
}

/**
 *    Calculation of the root mean square radius.
 *    rms  - root mean square radius
 *    psi  - array with the wave function values
 *    tmpz - temporary array
 */
void calcrms(double *rms, double *psi, double *tmpz) {
   long cnti;
   double psi2;

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nz; cnti ++) {
      psi2 = psi[cnti] * psi[cnti];
      tmpz[cnti] = z2[cnti] * psi2;
   }

   *rms = sqrt(simpint_th(dz, tmpz, Nz));

   return;
}

/**
 *    Time propagation with respect to H1 (part of the Hamiltonian without
 *    spatial derivatives).
 *    psi    - array with the wave function values
 *    psidd2 - array with the squared wave function values
 */
void calcnu(double *psi, double *psidd2) {
   long cnti;
   double psi2, psi2lin, psidd2lin, tmp;

   calcpsidd2(psi, psidd2);

   #pragma omp parallel for private(cnti, psi2, psi2lin, psidd2lin, tmp)
   for(cnti = 0; cnti < Nz; cnti ++) {
      psi2 = psi[cnti] * psi[cnti];
      psi2lin = psi2 * g;
      psidd2lin = psidd2[cnti] * gd;
      tmp = dt * (pot[cnti] + psi2lin + psidd2lin);
      psi[cnti] *= exp(- tmp);
   }

   return;
}

/**
 *    Time propagation with respect to H2 (z-part of the Laplacian).
 *    psi   - array with the wave function values
 *    cbeta - Crank-Nicolson scheme coefficients
 */
void calcluz(double *psi, double *cbeta) {
   long cnti;
   double c;

   cbeta[Nz - 2] = psi[Nz - 1];
   for (cnti = Nz - 2; cnti > 0; cnti --) {
      c = - Az * psi[cnti + 1] + Az0r * psi[cnti] - Az * psi[cnti - 1];
      cbeta[cnti - 1] =  cgammaz[cnti] * (Az * cbeta[cnti] - c);
   }
   psi[0] = 0.;
   for (cnti = 0; cnti < Nz - 2; cnti ++) {
      psi[cnti + 1] = calphaz[cnti] * psi[cnti] + cbeta[cnti];
   }
   psi[Nz - 1] = 0.;

   return;
}

void outpsi2z(double *psi, FILE *file) {
   long cnti;

   for(cnti = 0; cnti < Nz; cnti += outstpz) {
      fprintf(file, "%8le %8le\n", z[cnti], psi[cnti] * psi[cnti]);
   }
   fflush(file);

   return;
}
