/**
 * DBEC-GP-OMP-CUDA-MPI programs are developed by:
 *
 * Vladimir Loncar, Antun Balaz
 * (Scientific Computing Laboratory, Institute of Physics Belgrade, Serbia)
 *
 * Srdjan Skrbic
 * (Department of Mathematics and Informatics, Faculty of Sciences, University of Novi Sad, Serbia)
 *
 * Paulsamy Muruganandam
 * (Bharathidasan University, Tamil Nadu, India)
 *
 * Luis E. Young-S, Sadhan K. Adhikari
 * (UNESP - Sao Paulo State University, Brazil)
 *
 *
 * Public use and modification of these codes are allowed provided that the
 * following papers are cited:
 * [1] V. Loncar et al., Comput. Phys. Commun. 209 (2016) 190.      
 * [2] V. Loncar et al., Comput. Phys. Commun. 200 (2016) 406.      
 * [3] R. Kishor Kumar et al., Comput. Phys. Commun. 195 (2015) 117.
 *
 * The authors would be grateful for all information and/or comments
 * regarding the use of the programs.
 */

#include "imag2dXZ-th.h"

int main(int argc, char **argv) {
   FILE *out;
   FILE *file;
   FILE *filerms;
   char filename[MAX_FILENAME_SIZE];
   int nthreads;
   long cnti;
   double norm, mu, en;
   double *rms;
   double **cbeta;
   double **dpsi;
   double **tmpzi, **tmpxi, **tmpzj, **tmpxj;
   double psi2;
   fftw_complex *psidd2fft;

   pi = 3.14159265358979;

   struct timeval start, stop, iter_start, iter_stop;
   double wall_time, init_time, iter_time;
   iter_time = 0.;
   gettimeofday(&start, NULL);

   if ((argc != 3) || (strcmp(*(argv + 1), "-i") != 0)) {
      fprintf(stderr, "Usage: %s -i <input parameter file> \n", *argv);
      exit(EXIT_FAILURE);
   }

   if (! cfg_init(argv[2])) {
      fprintf(stderr, "Wrong input parameter file.\n");
      exit(EXIT_FAILURE);
   }

   readpar();

   #pragma omp parallel
   #pragma omp master
   nthreads = omp_get_num_threads();

   rms = alloc_double_vector(RMS_ARRAY_SIZE);

   z = alloc_double_vector(Nz);
   x = alloc_double_vector(Nx);

   z2 = alloc_double_vector(Nz);
   x2 = alloc_double_vector(Nx);

   pot = alloc_double_matrix(Nz, Nx);
   potdd = alloc_double_matrix(Nz, Nx);
   psi = alloc_double_matrix(Nz, Nx);
   psidd2 = alloc_double_matrix(Nz, (Nx / 2 + 1) * 2);
   psidd2fft = (fftw_complex *) *psidd2;

   dpsi = alloc_double_matrix(Nz, Nx);

   calphaz = alloc_double_vector(Nz - 1);
   calphax = alloc_double_vector(Nx - 1);
   cbeta =  alloc_double_matrix(nthreads, MAX(Nz, Nx) - 1);
   cgammaz = alloc_double_vector(Nz - 1);
   cgammax = alloc_double_vector(Nx - 1);

   tmpzi = alloc_double_matrix(nthreads, Nz);
   tmpxi = alloc_double_matrix(nthreads, Nx);
   tmpzj = alloc_double_matrix(nthreads, Nz);
   tmpxj = alloc_double_matrix(nthreads, Nx);

   if (output != NULL) {
      sprintf(filename, "%s.txt", output);
      out = fopen(filename, "w");
   } else out = stdout;

   if (rmsout != NULL) {
      sprintf(filename, "%s.txt", rmsout);
      filerms = fopen(filename, "w");
   } else filerms = NULL;

   if (opt == 2) par = 2.;
   else par = 1.;

   g_2d = g0 / (sqrt(2. * pi)  * d_y);
   gd_2d = gd0 / (sqrt(2. * pi) * d_y);

   Nz2 = Nz / 2; Nx2 = Nx / 2;
   dz2 = dz * dz; dx2 = dx * dx;

   fprintf(out, " Imaginary time propagation 2dXZ,   OPTION = %d, OMP_NUM_THREADS = %d\n\n", opt, nthreads);
   fprintf(out, "  Number of Atoms N = %li, Unit of length AHO = %.8f m\n", Na, aho);
   fprintf(out, "  Scattering length a = %.2f*a0, Dipolar ADD = %.2f*a0\n", as, add);
   fprintf(out, "  Nonlinearity G_3D = %.4f, Strength of DDI GD_3D = %.5f\n", g0, gd0);
   fprintf(out, "  Nonlinearity G_2D = %.4f, Strength of DDI GD_2D = %.5f\n", g_2d, gd_2d);
   fprintf(out, "  Parameters of trap: GAMMA = %.2f, LAMBDA = %.2f\n", vgamma, vlambda);
   fprintf(out, "  Axial Gaussian Width = %.2f\n\n", d_y);
   fprintf(out, " # Space Stp: NX = %li, NZ = %li\n", Nx, Nz);
   fprintf(out, "  Space Step: DX = %.6f, DZ = %.6f\n", dx, dz);
   fprintf(out, " # Time Stp : NSTP = %li, NPAS = %li, NRUN = %li\n", Nstp, Npas, Nrun);
   fprintf(out, "   Time Step:   DT = %.6f\n\n",  dt);
   fprintf(out, "                  --------------------------------------------------------\n");
   fprintf(out, "                    Norm      Chem        Ener/N     <rho>     |Psi(0,0)|^2\n");
   fprintf(out, "                  --------------------------------------------------------\n");
   fflush(out);

   g = 0.;
   gd = 0.;

   fftw_init_threads();
   fftw_plan_with_nthreads(nthreads);
   initpsi(psi);
   initpot();
   gencoef();
   initpotdd(*tmpzi, *tmpxi, *tmpzj, *tmpxj);

   plan_forward = fftw_plan_dft_r2c_2d(Nz, Nx, *psidd2, psidd2fft, FFT_FLAG);
   plan_backward = fftw_plan_dft_c2r_2d(Nz, Nx, psidd2fft, *psidd2, FFT_FLAG);

   calcnorm(&norm, psi, tmpzi, tmpxi);
   calcmuen(&mu, &en, psi, dpsi, psidd2, tmpzi, tmpxi, tmpzj, tmpxj);
   calcrms(rms, psi, tmpzi, tmpxi);

   psi2 = psi[Nz2][Nx2] * psi[Nz2][Nx2];
   fprintf(out, "Initial : %15.4f %11.5f %11.5f %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
   fflush(out);
   if (initout != NULL) {
      sprintf(filename, "%s.txt", initout);
      file = fopen(filename, "w");
      outpsi2xz(psi, file);
      fclose(file);

      sprintf(filename, "%s1d_z.txt", initout);
      file = fopen(filename, "w");
      outdenz(psi, *tmpxi, file);
      fclose(file);

      sprintf(filename, "%s1d_x.txt", initout);
      file = fopen(filename, "w");
      outdenx(psi, *tmpzi, file);
      fclose(file);
   }
   if (rmsout != NULL) {
      fprintf(filerms, " Imaginary time propagation 2dXZ,   OPTION = %d\n\n", opt);
      fprintf(filerms, "                  --------------------------------------------------------\n");
      fprintf(filerms, "Values of rms size:        <z>             <x>             <rho>\n");
      fprintf(filerms, "                  --------------------------------------------------------\n");
      fprintf(filerms, "           Initial:%14.5f   %13.5f   %13.5f\n", rms[1], rms[2], rms[0]);
      fflush(filerms);
   }

   if (Nstp != 0) {
      double g_stp = par * g_2d / (double) Nstp;
      double gd_stp = par * gd0 / (double) Nstp;
      g = 0.;
      gd = 0.;
      gettimeofday(&iter_start, NULL);
      for (cnti = 0; cnti < Nstp; cnti ++) {
         g += g_stp;
         gd += gd_stp;
         calcnu(psi, psidd2);
         calcluz(psi, cbeta);
         calclux(psi, cbeta);
         calcnorm(&norm, psi, tmpzi, tmpxi);
      }
      gettimeofday(&iter_stop, NULL);
      iter_time += (double) (((iter_stop.tv_sec - iter_start.tv_sec) * 1000 + (iter_stop.tv_usec - iter_start.tv_usec)/1000.0) + 0.5);

      calcmuen(&mu, &en, psi, dpsi, psidd2, tmpzi, tmpxi, tmpzj, tmpxj);
      calcrms(rms, psi, tmpzi, tmpxi);
      psi2 = psi[Nz2][Nx2] * psi[Nz2][Nx2];
      fprintf(out, "After NSTP iter.:%8.4f %11.5f %11.5f %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
      fflush(out);
      if (Nstpout != NULL) {
         sprintf(filename, "%s.txt", Nstpout);
         file = fopen(filename, "w");
         outpsi2xz(psi, file);
         fclose(file);

         sprintf(filename, "%s1d_z.txt", Nstpout);
         file = fopen(filename, "w");
         outdenz(psi, *tmpxi, file);
         fclose(file);

         sprintf(filename, "%s1d_x.txt", Nstpout);
         file = fopen(filename, "w");
         outdenx(psi, *tmpzi, file);
         fclose(file);
      }
      if (rmsout != NULL) {
         fprintf(filerms, "  After NSTP iter.:%14.5f   %13.5f   %13.5f\n", rms[1], rms[2], rms[0]);
         fflush(filerms);
      }
   } else {
      g = par * g_2d;
      gd = par * gd0;
   }

   gettimeofday(&iter_start, NULL);
   for (cnti = 0; cnti < Npas; cnti ++) {
      calcnu(psi, psidd2);
      calcluz(psi, cbeta);
      calclux(psi, cbeta);
      calcnorm(&norm, psi, tmpzi, tmpxi);
   }
   gettimeofday(&iter_stop, NULL);
   iter_time += (double) (((iter_stop.tv_sec - iter_start.tv_sec) * 1000 + (iter_stop.tv_usec - iter_start.tv_usec)/1000.0) + 0.5);

   calcmuen(&mu, &en, psi, dpsi, psidd2, tmpzi, tmpxi, tmpzj, tmpxj);
   calcrms(rms, psi, tmpzi, tmpxi);
   psi2 = psi[Nz2][Nx2] * psi[Nz2][Nx2];
   fprintf(out, "After NPAS iter.:%8.4f %11.5f %11.5f %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
   fflush(out);
   if (Npasout != NULL) {
      sprintf(filename, "%s.txt", Npasout);
      file = fopen(filename, "w");
      outpsi2xz(psi, file);
      fclose(file);

      sprintf(filename, "%s1d_z.txt", Npasout);
      file = fopen(filename, "w");
      outdenz(psi, *tmpxi, file);
      fclose(file);

      sprintf(filename, "%s1d_x.txt", Npasout);
      file = fopen(filename, "w");
      outdenx(psi, *tmpzi, file);
      fclose(file);
   }
   if (rmsout != NULL) {
      fprintf(filerms, "  After NPAS iter.:%14.5f   %13.5f   %13.5f\n", rms[1], rms[2], rms[0]);
      fflush(filerms);
   }

   gettimeofday(&iter_start, NULL);
   for (cnti = 0; cnti < Nrun; cnti ++) {
      calcnu(psi, psidd2);
      calcluz(psi, cbeta);
      calclux(psi, cbeta);
      calcnorm(&norm, psi, tmpzi, tmpxi);
   }
   gettimeofday(&iter_stop, NULL);
   iter_time += (double) (((iter_stop.tv_sec - iter_start.tv_sec) * 1000 + (iter_stop.tv_usec - iter_start.tv_usec)/1000.0) + 0.5);

   calcmuen(&mu, &en, psi, dpsi, psidd2, tmpzi, tmpxi, tmpzj, tmpxj);
   calcrms(rms, psi, tmpzi, tmpxi);
   psi2 = psi[Nz2][Nx2] * psi[Nz2][Nx2];
   fprintf(out, "After NRUN iter.:%8.4f %11.5f %11.5f %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
   fflush(out);
   if (Nrunout != NULL) {
      sprintf(filename, "%s.txt", Nrunout);
      file = fopen(filename, "w");
      outpsi2xz(psi, file);
      fclose(file);

      sprintf(filename, "%s1d_z.txt", Nrunout);
      file = fopen(filename, "w");
      outdenz(psi, *tmpxi, file);
      fclose(file);

      sprintf(filename, "%s1d_x.txt", Nrunout);
      file = fopen(filename, "w");
      outdenx(psi, *tmpzi, file);
      fclose(file);
   }
   if (rmsout != NULL) {
      fprintf(filerms, "  After NRUN iter.:%14.5f   %13.5f   %13.5f\n", rms[1], rms[2], rms[0]);
      fprintf(filerms, "                  --------------------------------------------------------\n");
   }
   if (rmsout != NULL) fclose(filerms);

   fprintf(out, "                  --------------------------------------------------------\n\n");

   free_double_vector(rms);

   free_double_vector(z);
   free_double_vector(x);

   free_double_vector(z2);
   free_double_vector(x2);

   free_double_matrix(pot);
   free_double_matrix(potdd);
   free_double_matrix(psi);
   free_double_matrix(psidd2);

   free_double_matrix(dpsi);

   free_double_vector(calphaz);
   free_double_vector(calphax);
   free_double_matrix(cbeta);
   free_double_vector(cgammaz);
   free_double_vector(cgammax);

   free_double_matrix(tmpzi);
   free_double_matrix(tmpxi);
   free_double_matrix(tmpzj);
   free_double_matrix(tmpxj);

   fftw_destroy_plan(plan_forward);
   fftw_destroy_plan(plan_backward);

   gettimeofday(&stop, NULL);
   wall_time = (double) (((stop.tv_sec - start.tv_sec) * 1000 + (stop.tv_usec - start.tv_usec)/1000.0) + 0.5);
   init_time = wall_time - iter_time;
   fprintf(out, " Initialization/allocation wall-clock time: %.3f seconds\n", init_time / 1000.);
   fprintf(out, " Calculation (iterations) wall-clock time: %.3f seconds\n", iter_time / 1000.);

   if(output != NULL) fclose(out);

   return (EXIT_SUCCESS);
}

/**
 *    Reading input parameters from the configuration file.
 */
void readpar(void) {
   char *cfg_tmp;

   if ((cfg_tmp = cfg_read("OPTION")) == NULL) {
      fprintf(stderr, "OPTION is not defined in the configuration file\n");
      exit(EXIT_FAILURE);
   }
   opt = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("NATOMS")) == NULL) {
      fprintf(stderr, "NATOMS is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Na = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("G0")) == NULL) {
      if ((cfg_tmp = cfg_read("AHO")) == NULL) {
         fprintf(stderr, "AHO is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      aho = atof(cfg_tmp);

      if ((cfg_tmp = cfg_read("AS")) == NULL) {
         fprintf(stderr, "AS is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      as = atof(cfg_tmp);

      g0 = 4. * pi * as * Na * BOHR_RADIUS / aho;
   } else {
      g0 = atof(cfg_tmp);
   }

   if ((cfg_tmp = cfg_read("GDD0")) == NULL) {
      if ((cfg_tmp = cfg_read("AHO")) == NULL) {
         fprintf(stderr, "AHO is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      aho = atof(cfg_tmp);

      if ((cfg_tmp = cfg_read("ADD")) == NULL) {
         fprintf(stderr, "ADD is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      add = atof(cfg_tmp);

      gd0 = 3. * add * Na * BOHR_RADIUS / aho;
   } else {
      gd0 = atof(cfg_tmp);
   }

   if ((cfg_tmp = cfg_read("NZ")) == NULL) {
      fprintf(stderr, "NZ is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Nz = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("NX")) == NULL) {
      fprintf(stderr, "NX is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Nx = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("DZ")) == NULL) {
      fprintf(stderr, "DZ is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   dz = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("DX")) == NULL) {
      fprintf(stderr, "DX is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   dx = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("DT")) == NULL) {
      fprintf(stderr, "DT is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   dt = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("LAMBDA")) == NULL) {
      fprintf(stderr, "LAMBDA is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   vlambda = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("GAMMA")) == NULL) {
      fprintf(stderr, "GAMMA is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   vgamma = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("D_Y")) == NULL) {
      fprintf(stderr, "D_Y is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   d_y = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("NSTP")) == NULL) {
      fprintf(stderr, "NSTP is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Nstp = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("NPAS")) == NULL) {
      fprintf(stderr, "NPAS is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Npas = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("NRUN")) == NULL) {
      fprintf(stderr, "NRUN is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Nrun = atol(cfg_tmp);

   output = cfg_read("OUTPUT");
   rmsout = cfg_read("RMSOUT");
   initout = cfg_read("INITOUT");
   Nstpout = cfg_read("NSTPOUT");
   Npasout = cfg_read("NPASOUT");
   Nrunout = cfg_read("NRUNOUT");

   if((initout != NULL) || (Nstpout != NULL)  || (Npasout != NULL) || (Nrunout != NULL)) {
      if((cfg_tmp = cfg_read("OUTSTPZ")) == NULL) {
         fprintf(stderr, "OUTSTPZ is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      outstpz = atol(cfg_tmp);

      if((cfg_tmp = cfg_read("OUTSTPX")) == NULL) {
         fprintf(stderr, "OUTSTPX is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      outstpx = atol(cfg_tmp);
   }

   return;
}

/**
 *    Initialization of the space mesh and the initial wave function.
 *    psi - array with the wave function values
 */
void initpsi(double **psi) {
   long cnti, cntj;
   double cpsi;
   double tmp;

   cpsi = sqrt(pi * sqrt(1. / (vgamma * vlambda)));

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nz; cnti ++) {
      z[cnti] = (cnti - Nz2) * dz;
      z2[cnti] = z[cnti] * z[cnti];
   }

   #pragma omp parallel for private(cntj)
   for(cntj = 0; cntj < Nx; cntj ++) {
      x[cntj] = (cntj - Nx2) * dx;
      x2[cntj] = x[cntj] * x[cntj];
   }

   #pragma omp parallel for private(cnti, cntj, tmp)
   for(cnti = 0; cnti < Nz; cnti ++) {
      for(cntj = 0; cntj < Nx; cntj ++) {
         tmp = exp(- 0.5 * (vgamma * x2[cntj] + vlambda * z2[cnti]));
         psi[cnti][cntj] = tmp / cpsi;
      }
   }

   return;
}

/**
 *    Initialization of the potential.
 */
void initpot() {
   long cnti, cntj;
   double vlambda2, vgamma2;

   vlambda2 = vlambda * vlambda;
   vgamma2 = vgamma * vgamma;

   #pragma omp parallel for private(cnti, cntj)
   for(cnti = 0; cnti < Nz; cnti ++) {
      for(cntj = 0; cntj < Nx; cntj ++) {
         pot[cnti][cntj] = 0.5 * par * (vgamma2 * x2[cntj] + vlambda2 * z2[cnti]);
      }
   }

   return;
}

/**
 *    Crank-Nicolson scheme coefficients generation.
 */
void gencoef(void) {
   long cnti;

   Az0 = 1. + dt / dz2 / (3. - par);
   Ax0 = 1. + dt / dx2 / (3. - par);

   Az0r = 1. - dt / dz2 / (3. - par);
   Ax0r = 1. - dt / dx2 / (3. - par);

   Az = - 0.5 * dt / dz2 / (3. - par);
   Ax = - 0.5 * dt / dx2 / (3. - par);

   calphaz[Nz - 2] = 0.;
   cgammaz[Nz - 2] = - 1. / Az0;
   for(cnti = Nz - 2; cnti > 0; cnti --) {
      calphaz[cnti - 1] = Az * cgammaz[cnti];
      cgammaz[cnti - 1] = - 1. / (Az0 + Az * calphaz[cnti - 1]);
   }

   calphax[Nx - 2] = 0.;
   cgammax[Nx - 2] = - 1. / Ax0;
   for(cnti = Nx - 2; cnti > 0; cnti --) {
      calphax[cnti - 1] = Ax * cgammax[cnti];
      cgammax[cnti - 1] = - 1. / (Ax0 + Ax * calphax[cnti - 1]);
   }

   return;
}
/**
 *    Initialization of the dipolar potential.
 *    kz  - array with the space mesh values in the z-direction in the K-space
 *    kx  - array with the space mesh values in the x-direction in the K-space
 *    kz2 - array with the squared space mesh values in the z-direction in the
 *          K-space
 *    kx2 - array with the squared space mesh values in the x-direction in the
 *          K-space
 */
void initpotdd(double *kz, double *kx, double *kz2, double *kx2) {
   long cnti, cntj;
   double dkz, dkx;
   double zi, zi2;

   dkz = 2. * pi / (Nz * dz);
   dkx = 2. * pi / (Nx * dx);

   for(cnti = 0; cnti < Nz2; cnti ++) kz[cnti] = cnti * dkz;
   for(cnti = 0; cnti < Nz2; cnti ++) kz[cnti + Nz2] = (cnti - Nz2) * dkz;
   for(cntj = 0; cntj < Nx2; cntj ++) kx[cntj] = cntj * dkx;
   for(cntj = 0; cntj < Nx2; cntj ++) kx[cntj + Nx2] = (cntj - Nx2) * dkx;

   for(cnti = 0; cnti < Nz; cnti ++) kz2[cnti] = kz[cnti] * kz[cnti];
   for(cntj = 0; cntj < Nx; cntj ++) kx2[cntj] = kx[cntj] * kx[cntj];

   #pragma omp parallel for private(cnti, cntj, zi, zi2)
   for(cnti = 0; cnti < Nz; cnti ++) {
      for(cntj = 0; cntj < Nx; cntj ++) {
         zi2 = (kz2[cnti] + kx2[cntj]) * d_y * d_y / 2.;
         zi = sqrt(zi2);

         if(zi >= XI_LIMIT) {
            potdd[cnti][cntj] = -1. + 3. * (kz2[cnti] * d_y * d_y / 2.) / zi2;
         } else if((fabs(kz[cnti]) < 0.01) && (fabs(kx[cntj]) < 0.01)) {
            potdd[cnti][cntj] = -1.;
         } else {
            potdd[cnti][cntj] = -1. + 3. * sqrt(pi) * (kz2[cnti] * d_y * d_y / 2.) / zi * exp(zi2) * erfc(zi);
         }
         potdd[cnti][cntj] *= (4. * pi / 3.) / (sqrt(2. * pi) * d_y);
      }
   }

   return;
}

/**
 *    Calculation of the wave function norm and normalization.
 *    norm - wave function norm
 *    psi  - array with the wave function values
 *    tmpz - temporary array
 *    tmpx - temporary array
 */
void calcnorm(double *norm, double **psi, double **tmpz, double **tmpx) {
   int threadid;
   long cnti, cntj;
   double tmp;

   #pragma omp parallel private(threadid, cnti, cntj)
   {
      threadid = omp_get_thread_num();

      #pragma omp for
      for (cnti = 0; cnti < Nz; cnti ++) {
         for (cntj = 0; cntj < Nx; cntj ++) {
            tmpx[threadid][cntj] = psi[cnti][cntj] * psi[cnti][cntj];
         }
         (*tmpz)[cnti] = simpint(dx, tmpx[threadid], Nx);
      }
      #pragma omp barrier

      #pragma omp single
      {
         *norm = sqrt(simpint(dz, *tmpz, Nz));
         tmp = 1. / *norm;
      }

      #pragma omp for
      for (cnti = 0; cnti < Nz; cnti ++) {
         for (cntj = 0; cntj < Nx; cntj ++) {
            psi[cnti][cntj] *= tmp;
         }
      }
   }

   return;
}

/**
 *    Calculation of the chemical potential and energy.
 *    mu     - chemical potential
 *    en     - energy
 *    psi    - array with the wave function values
 *    dpsi   - temporary array
 *    psidd2 - array with the squared wave function values
 *    tmpzi  - temporary array
 *    tmpxi  - temporary array
 *    tmpzj  - temporary array
 *    tmpxj  - temporary array
 */
void calcmuen(double *mu, double *en, double **psi, double **dpsi, double **psidd2, double **tmpzi, double **tmpxi, double **tmpzj, double **tmpxj) {
   int threadid;
   long cnti, cntj;
   double psi2, psi2lin, psidd2lin, dpsi2;

   calcpsidd2(psi, psidd2);

   #pragma omp parallel private(threadid, cnti, cntj, psi2, psi2lin, psidd2lin, dpsi2)
   {
      threadid = omp_get_thread_num();

      #pragma omp for
      for (cntj = 0; cntj < Nx; cntj ++) {
         for (cnti = 0; cnti < Nz; cnti ++) {
            tmpzi[threadid][cnti] = psi[cnti][cntj];
         }
         diff(dz, tmpzi[threadid], tmpzj[threadid], Nz);
         for (cnti = 0; cnti < Nz; cnti ++) {
            dpsi[cnti][cntj] = tmpzj[threadid][cnti] * tmpzj[threadid][cnti];
         }
      }

      #pragma omp for
      for (cnti = 0; cnti < Nz; cnti ++) {
         for (cntj = 0; cntj < Nx; cntj ++) {
            tmpxi[threadid][cntj] = psi[cnti][cntj];
         }
         diff(dx, tmpxi[threadid], tmpxj[threadid], Nx);
         for (cntj = 0; cntj < Nx; cntj ++) {
            dpsi[cnti][cntj] += tmpxj[threadid][cntj] * tmpxj[threadid][cntj];
         }
      }

      #pragma omp barrier

      #pragma omp for
      for (cnti = 0; cnti < Nz; cnti ++) {
         for (cntj = 0; cntj < Nx; cntj ++) {
            psi2 = psi[cnti][cntj] * psi[cnti][cntj];
            psi2lin = psi2 * g;
            psidd2lin = psidd2[cnti][cntj] * gd;
            dpsi2 = dpsi[cnti][cntj] / (3. - par);
            tmpxi[threadid][cntj] = (pot[cnti][cntj] + psi2lin + psidd2lin) * psi2 + dpsi2;
            tmpxj[threadid][cntj] = (pot[cnti][cntj] + 0.5 * psi2lin + 0.5 * psidd2lin) * psi2 + dpsi2;
         }
         (*tmpzi)[cnti] = simpint(dx, tmpxi[threadid], Nx);
         (*tmpzj)[cnti] = simpint(dx, tmpxj[threadid], Nx);
      }
   }

   *mu = simpint(dz, *tmpzi, Nz);
   *en = simpint(dz, *tmpzj, Nz);

   return;
}

/**
 *    Calculation of squared wave function values for dipole-dipole
 *    interaction.
 *    psi    - array with the wave function values
 *    psidd2 - array with the squared wave function values
 */
void calcpsidd2(double **psi, double **psidd2) {
   long cnti, cntj;
   fftw_complex *psidd2fft = (fftw_complex *) *psidd2;

   #pragma omp parallel for private(cnti, cntj)
   for (cnti = 0; cnti < Nz; cnti ++) {
      for (cntj = 0; cntj < Nx; cntj ++) {
         psidd2[cnti][cntj] = psi[cnti][cntj] * psi[cnti][cntj];
      }
   }

   fftw_execute_dft_r2c(plan_forward, *psidd2, psidd2fft);

   #pragma omp parallel for private(cnti, cntj)
   for (cnti = 0; cnti < Nz; cnti ++) {
      for (cntj = 0; cntj < (Nx / 2) + 1; cntj ++) {
         psidd2fft[cnti * (Nx2 + 1) + cntj][0] *= potdd[cnti][cntj];
         psidd2fft[cnti * (Nx2 + 1) + cntj][1] *= potdd[cnti][cntj];
      }
   }

   fftw_execute_dft_c2r(plan_backward, psidd2fft, *psidd2);

   #pragma omp parallel for private(cnti, cntj)
   for (cnti = 0; cnti < Nz; cnti ++) {
      for (cntj = 0; cntj < Nx; cntj ++) {
         psidd2[cnti][cntj] /= Nz * Nx;
      }
   }

   #pragma omp parallel for private(cntj)
   for (cntj = 0; cntj < Nx - 1; cntj ++) {
      psidd2[Nz - 1][cntj] = psidd2[0][cntj];
   }

   #pragma omp parallel for private(cnti)
   for (cnti = 0; cnti < Nz - 1; cnti ++) {
      psidd2[cnti][Nx - 1] = psidd2[cnti][0];
   }

   return;
}

/**
 *    Calculation of the root mean square radius.
 *    rms  - root mean square radius
 *    psi  - array with the wave function values
 *    tmpz - temporary array
 *    tmpx - temporary array
 */
void calcrms(double *rms, double **psi, double **tmpz, double **tmpx) {
   int threadid;
   long cnti, cntj;
   double psi2;

   #pragma omp parallel private(threadid, cnti, cntj, psi2)
   {
      threadid = omp_get_thread_num();

      #pragma omp for
      for (cntj = 0; cntj < Nx; cntj ++) {
         for (cnti = 0; cnti < Nz; cnti ++) {
            psi2 = psi[cnti][cntj] * psi[cnti][cntj];
            tmpz[threadid][cnti] = z2[cnti] * psi2;
         }
         (*tmpx)[cntj] = simpint(dz, tmpz[threadid], Nz);
      }
      #pragma omp barrier

      #pragma omp single
      rms[1] = sqrt(simpint(dx, *tmpx, Nx));

      #pragma omp for
      for (cnti = 0; cnti < Nz; cnti ++) {
         for (cntj = 0; cntj < Nx; cntj ++) {
            psi2 = psi[cnti][cntj] * psi[cnti][cntj];
            tmpx[threadid][cntj] = x2[cntj] * psi2;
         }
         (*tmpz)[cnti] = simpint(dx, tmpx[threadid], Nx);
      }
      #pragma omp barrier

      #pragma omp single
      rms[2] = sqrt(simpint(dz, *tmpz, Nz));
   }

   rms[0] = sqrt(rms[1] * rms[1] + rms[2] * rms[2]);

   return;
}

/**
 *    Time propagation with respect to H1 (part of the Hamiltonian without
 *    spatial derivatives).
 *    psi    - array with the wave function values
 *    psidd2 - array with the squared wave function values
 */
void calcnu(double **psi, double **psidd2) {
   long cnti, cntj;
   double psi2, psi2lin, psidd2lin, tmp;

   calcpsidd2(psi, psidd2);

   #pragma omp parallel for private(cnti, cntj, psi2, psi2lin, psidd2lin, tmp)
   for (cnti = 0; cnti < Nz; cnti ++) {
      for (cntj = 0; cntj < Nx; cntj ++) {
         psi2 = psi[cnti][cntj] * psi[cnti][cntj];
         psi2lin = psi2 * g;
         psidd2lin = psidd2[cnti][cntj] * gd;
         tmp = dt * (pot[cnti][cntj] + psi2lin + psidd2lin);
         psi[cnti][cntj] *= exp(- tmp);
      }
   }

   return;
}

/**
 *    Time propagation with respect to H2 (z-part of the Laplacian).
 *    psi   - array with the wave function values
 *    cbeta - Crank-Nicolson scheme coefficients
 */
void calcluz(double **psi, double **cbeta) {
   int threadid;
   long cnti, cntj;
   double c;

   #pragma omp parallel private(threadid, cnti, cntj, c)
   {
      threadid = omp_get_thread_num();

      #pragma omp for
      for (cntj = 0; cntj < Nx; cntj ++) {
         cbeta[threadid][Nz - 2] = psi[Nz - 1][cntj];
         for (cnti = Nz - 2; cnti > 0; cnti --) {
            c = - Az * psi[cnti + 1][cntj] + Az0r * psi[cnti][cntj] - Az * psi[cnti - 1][cntj];
            cbeta[threadid][cnti - 1] =  cgammaz[cnti] * (Az * cbeta[threadid][cnti] - c);
         }
         psi[0][cntj] = 0.;
         for (cnti = 0; cnti < Nz - 2; cnti ++) {
            psi[cnti + 1][cntj] = calphaz[cnti] * psi[cnti][cntj] + cbeta[threadid][cnti];
         }
         psi[Nz - 1][cntj] = 0.;
      }
   }

   return;
}

/**
 *    Time propagation with respect to H3 (x-part of the Laplacian).
 *    psi   - array with the wave function values
 *    cbeta - Crank-Nicolson scheme coefficients
 */
void calclux(double **psi, double **cbeta) {
   int threadid;
   long cnti, cntj;
   double c;

   #pragma omp parallel private(threadid, cnti, cntj, c)
   {
      threadid = omp_get_thread_num();

      #pragma omp for
      for (cnti = 0; cnti < Nz; cnti ++) {
         cbeta[threadid][Nx - 2] = psi[cnti][Nx - 1];
         for (cntj = Nx - 2; cntj > 0; cntj --) {
            c = - Ax * psi[cnti][cntj + 1] + Ax0r * psi[cnti][cntj] - Ax * psi[cnti][cntj - 1];
            cbeta[threadid][cntj - 1] =  cgammax[cntj] * (Ax * cbeta[threadid][cntj] - c);
         }
         psi[cnti][0] = 0.;
         for (cntj = 0; cntj < Nx - 2; cntj ++) {
            psi[cnti][cntj + 1] = calphax[cntj] * psi[cnti][cntj] + cbeta[threadid][cntj];
         }
         psi[cnti][Nx - 1] = 0.;
      }
   }

   return;
}

void outpsi2xz(double **psi, FILE *file) {
   long cnti, cntj;

   for(cnti = 0; cnti < Nz; cnti += outstpz) {
      for(cntj = 0; cntj < Nx; cntj += outstpx) {
         fprintf(file, "%8le %8le %8le\n", z[cnti], x[cntj], psi[cnti][cntj] * psi[cnti][cntj]);
      }
      fprintf(file, "\n");
      fflush(file);
   }

   return;
}

void outdenz(double **psi, double *tmp, FILE *file) {
   long cnti, cntj;

   for(cnti = 0; cnti < Nz; cnti += outstpz) {
      for(cntj = 0; cntj < Nx; cntj ++) {
            tmp[cntj] = psi[cnti][cntj] * psi[cnti][cntj];
      }
      fprintf(file, "%8le %8le\n", z[cnti], simpint(dx, tmp, Nx));
      fflush(file);
   }
}

void outdenx(double **psi, double *tmp, FILE *file) {
   long cnti, cntj;

   for(cntj = 0; cntj < Nx; cntj += outstpx) {
       for(cnti = 0; cnti < Nz; cnti ++) {
            tmp[cnti] = psi[cnti][cntj] * psi[cnti][cntj];
      }
      fprintf(file, "%8le %8le\n", x[cntj], simpint(dz, tmp, Nz));
      fflush(file);
   }
}
