/**
 * DBEC-GP-OMP-CUDA-MPI programs are developed by:
 *
 * Vladimir Loncar, Antun Balaz
 * (Scientific Computing Laboratory, Institute of Physics Belgrade, Serbia)
 *
 * Srdjan Skrbic
 * (Department of Mathematics and Informatics, Faculty of Sciences, University of Novi Sad, Serbia)
 *
 * Paulsamy Muruganandam
 * (Bharathidasan University, Tamil Nadu, India)
 *
 * Luis E. Young-S, Sadhan K. Adhikari
 * (UNESP - Sao Paulo State University, Brazil)
 *
 *
 * Public use and modification of these codes are allowed provided that the
 * following papers are cited:
 * [1] V. Loncar et al., Comput. Phys. Commun. 209 (2016) 190.      
 * [2] V. Loncar et al., Comput. Phys. Commun. 200 (2016) 406.      
 * [3] R. Kishor Kumar et al., Comput. Phys. Commun. 195 (2015) 117.
 *
 * The authors would be grateful for all information and/or comments
 * regarding the use of the programs.
 */

#include "imag3d-th.h"

int main(int argc, char **argv) {
   FILE *out;
   FILE *file;
   FILE *filerms;
   char filename[MAX_FILENAME_SIZE];
   int nthreads;
   long cnti;
   double norm, mu, en;
   double *rms;
   double **cbeta;
   double ***dpsi;
   double **tmpxi, **tmpyi, **tmpzi, **tmpxj, **tmpyj, **tmpzj;
   double psi2;
   fftw_complex *psidd2fft;

   pi = 3.14159265358979;

   struct timeval start, stop, iter_start, iter_stop;
   double wall_time, init_time, iter_time;
   iter_time = 0.;
   gettimeofday(&start, NULL);

   if ((argc != 3) || (strcmp(*(argv + 1), "-i") != 0)) {
      fprintf(stderr, "Usage: %s -i <input parameter file> \n", *argv);
      exit(EXIT_FAILURE);
   }

   if (! cfg_init(argv[2])) {
      fprintf(stderr, "Wrong input parameter file.\n");
      exit(EXIT_FAILURE);
   }

   readpar();

   #pragma omp parallel
   #pragma omp master
   nthreads = omp_get_num_threads();

   rms = alloc_double_vector(RMS_ARRAY_SIZE);

   x = alloc_double_vector(Nx);
   y = alloc_double_vector(Ny);
   z = alloc_double_vector(Nz);

   x2 = alloc_double_vector(Nx);
   y2 = alloc_double_vector(Ny);
   z2 = alloc_double_vector(Nz);

   pot = alloc_double_tensor(Nx, Ny, Nz);
   potdd = alloc_double_tensor(Nx, Ny, Nz);
   psi = alloc_double_tensor(Nx, Ny, Nz);
   psidd2 = alloc_double_tensor(Nx, Ny, (Nz / 2 + 1) * 2);
   psidd2fft = (fftw_complex *) **psidd2;

   dpsi = alloc_double_tensor(Nx, Ny, Nz);

   calphax = alloc_double_vector(Nx - 1);
   calphay = alloc_double_vector(Ny - 1);
   calphaz = alloc_double_vector(Nz - 1);
   cbeta = alloc_double_matrix(nthreads, MAX(Nx, Ny, Nz) - 1);
   cgammax = alloc_double_vector(Nx - 1);
   cgammay = alloc_double_vector(Ny - 1);
   cgammaz = alloc_double_vector(Nz - 1);

   tmpxi = alloc_double_matrix(nthreads, Nx);
   tmpyi = alloc_double_matrix(nthreads, Ny);
   tmpzi = alloc_double_matrix(nthreads, Nz);
   tmpxj = alloc_double_matrix(nthreads, Nx);
   tmpyj = alloc_double_matrix(nthreads, Ny);
   tmpzj = alloc_double_matrix(nthreads, Nz);

   if (output != NULL) {
      sprintf(filename, "%s.txt", output);
      out = fopen(filename, "w");
   } else out = stdout;

   if (rmsout != NULL) {
      sprintf(filename, "%s.txt", rmsout);
      filerms = fopen(filename, "w");
   } else filerms = NULL;

   if (opt == 2) par = 2.;
   else par = 1.;

   Nx2 = Nx / 2; Ny2 = Ny / 2; Nz2 = Nz / 2;
   dx2 = dx * dx; dy2 = dy * dy; dz2 = dz * dz;

   fprintf(out, " Imaginary time propagation 3D,   OPTION = %d, OMP_NUM_THREADS = %d\n\n", opt, nthreads);
   fprintf(out, "  Number of Atoms N = %li, Unit of length AHO = %.8f m\n", Na, aho);
   fprintf(out, "  Scattering length a = %.2f*a0, Dipolar ADD = %.2f*a0\n", as, add);
   fprintf(out, "  Nonlinearity G_3D = %.4f, Strength of DDI GD_3D = %.5f\n", g0, gd0);
   fprintf(out, "  Parameters of trap: GAMMA = %.2f, NU = %.2f, LAMBDA = %.2f\n\n", vgamma, vnu, vlambda);
   fprintf(out, " # Space Stp: NX = %li, NY = %li, NZ = %li\n", Nx, Ny, Nz);
   fprintf(out, "  Space Step: DX = %.6f, DY = %.6f, DZ = %.6f\n", dx, dy, dz);
   fprintf(out, " # Time Stp : NSTP = %li, NPAS = %li, NRUN = %li\n", Nstp, Npas, Nrun);
   fprintf(out, "   Time Step:   DT = %.6f\n",  dt);
   fprintf(out, "   Dipolar Cut off:   R = %.3f\n\n",  cutoff);
   fprintf(out, "                  --------------------------------------------------------\n");
   fprintf(out, "                    Norm      Chem        Ener/N     <r>     |Psi(0,0,0)|^2\n");
   fprintf(out, "                  --------------------------------------------------------\n");
   fflush(out);

   fftw_init_threads();
   fftw_plan_with_nthreads(nthreads);
   initpsi(psi);
   initpot();
   gencoef();
   initpotdd(*tmpxi, *tmpyi, *tmpzi, *tmpxj, *tmpyj, *tmpzj);

   plan_forward = fftw_plan_dft_r2c_3d(Nx, Ny, Nz, **psidd2, psidd2fft, FFT_FLAG);
   plan_backward = fftw_plan_dft_c2r_3d(Nx, Ny, Nz, psidd2fft, **psidd2, FFT_FLAG);


   calcnorm(&norm, psi, tmpxi, tmpyi, tmpzi);
   calcmuen(&mu, &en, psi, dpsi, psidd2, tmpxi, tmpyi, tmpzi, tmpxj, tmpyj, tmpzj);
   calcrms(rms, psi, tmpxi, tmpyi, tmpzi);

   psi2 = psi[Nx2][Ny2][Nz2] * psi[Nx2][Ny2][Nz2];
   fprintf(out, "Initial : %15.4f %11.5f %11.5f %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
   fflush(out);
   if (initout != NULL) {
      sprintf(filename, "%s.txt", initout);
      file = fopen(filename, "w");
      outdenxyz(psi, file);
      fclose(file);

      sprintf(filename, "%s1d_x.txt", initout);
      file = fopen(filename, "w");
      outdenx(psi, *tmpyi, *tmpzi, file);
      fclose(file);

      sprintf(filename, "%s1d_y.txt", initout);
      file = fopen(filename, "w");
      outdeny(psi, *tmpxi, *tmpzi, file);
      fclose(file);

      sprintf(filename, "%s1d_z.txt", initout);
      file = fopen(filename, "w");
      outdenz(psi, *tmpxi, *tmpyi, file);
      fclose(file);
      /*
            sprintf(filename, "%s2d_xy.txt", initout);
            file = fopen(filename, "w");
            outdenxy(psi, *tmpzi, file);
            fclose(file);

            sprintf(filename, "%s2d_xz.txt", initout);
            file = fopen(filename, "w");
            outdenxz(psi, *tmpyi, file);
            fclose(file);

            sprintf(filename, "%s2d_yz.txt", initout);
            file = fopen(filename, "w");
            outdenyz(psi, *tmpxi, file);
            fclose(file);

            sprintf(filename, "%s3d_x0z.txt", initout);
            file = fopen(filename, "w");
            outpsi2xz(psi, file);
            fclose(file);

            sprintf(filename, "%s3d_xy0.txt", initout);
            file = fopen(filename, "w");
            outpsi2xy(psi, file);
            fclose(file);
      */
   }
   if (rmsout != NULL) {
      fprintf(filerms, " Imaginary time propagation 3D,   OPTION = %d\n\n", opt);
      fprintf(filerms, "                  --------------------------------------------------------\n");
      fprintf(filerms, "Values of rms size:     <r>          <x>          <y>          <z>\n");
      fprintf(filerms, "                  --------------------------------------------------------\n");
      fprintf(filerms, "           Initial:%12.5f %12.5f %12.5f %12.5f\n", rms[0], rms[1], rms[2], rms[3]);
      fflush(filerms);
   }

   if (Nstp != 0) {
      double g_stp = par * g0 / (double) Nstp;
      double gd_stp = par * gd0 / (double) Nstp;
      g = 0.;
      gd = 0.;
      gettimeofday(&iter_start, NULL);
      for (cnti = 0; cnti < Nstp; cnti ++) {
         g += g_stp;
         gd += gd_stp;
         calcnu(psi, psidd2);
         calcluy(psi, cbeta);
         calcluz(psi, cbeta);
         calclux(psi, cbeta);
         calcnorm(&norm, psi, tmpxi, tmpyi, tmpzi);
      }
      gettimeofday(&iter_stop, NULL);
      iter_time += (double) (((iter_stop.tv_sec - iter_start.tv_sec) * 1000 + (iter_stop.tv_usec - iter_start.tv_usec)/1000.0) + 0.5);

      calcmuen(&mu, &en, psi, dpsi, psidd2, tmpxi, tmpyi, tmpzi, tmpxj, tmpyj, tmpzj);
      calcrms(rms, psi, tmpxi, tmpyi, tmpzi);
      psi2 = psi[Nx2][Ny2][Nz2] * psi[Nx2][Ny2][Nz2];
      fprintf(out, "After NSTP iter.:%8.4f %11.5f %11.5f %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
      fflush(out);
      if (Nstpout != NULL) {
         sprintf(filename, "%s.txt", Nstpout);
         file = fopen(filename, "w");
         outpsi2xz(psi, file);
         fclose(file);
      }
      if (rmsout != NULL) {
         fprintf(filerms, "  After NSTP iter.:%12.5f %12.5f %12.5f %12.5f\n", rms[0], rms[1], rms[2], rms[3]);
         fflush(filerms);
      }
      if (Nstpout != NULL) {
         sprintf(filename, "%s.txt", Nstpout);
         file = fopen(filename, "w");
         outdenxyz(psi, file);
         fclose(file);

         sprintf(filename, "%s1d_x.txt", Nstpout);
         file = fopen(filename, "w");
         outdenx(psi, *tmpyi, *tmpzi, file);
         fclose(file);

         sprintf(filename, "%s1d_y.txt", Nstpout);
         file = fopen(filename, "w");
         outdeny(psi, *tmpxi, *tmpzi, file);
         fclose(file);

         sprintf(filename, "%s1d_z.txt", Nstpout);
         file = fopen(filename, "w");
         outdenz(psi, *tmpxi, *tmpyi, file);
         fclose(file);
         /*
               sprintf(filename, "%s2d_xy.txt", Nstpout);
               file = fopen(filename, "w");
               outdenxy(psi, *tmpzi, file);
               fclose(file);

               sprintf(filename, "%s2d_xz.txt", Nstpout);
               file = fopen(filename, "w");
               outdenxz(psi, *tmpyi, file);
               fclose(file);

               sprintf(filename, "%s2d_yz.txt", Nstpout);
               file = fopen(filename, "w");
               outdenyz(psi, *tmpxi, file);
               fclose(file);

               sprintf(filename, "%s3d_x0z.txt", Nstpout);
               file = fopen(filename, "w");
               outpsi2xz(psi, file);
               fclose(file);

               sprintf(filename, "%s3d_xy0.txt", Nstpout);
               file = fopen(filename, "w");
               outpsi2xy(psi, file);
               fclose(file);
         */
      }
   } else {
      g = par * g0;
      gd = par * gd0;
   }

   gettimeofday(&iter_start, NULL);
   for (cnti = 0; cnti < Npas; cnti ++) {
      calcnu(psi, psidd2);
      calclux(psi, cbeta);
      calcluy(psi, cbeta);
      calcluz(psi, cbeta);
      calcnorm(&norm, psi, tmpxi, tmpyi, tmpzi);
   }
   gettimeofday(&iter_stop, NULL);
   iter_time += (double) (((iter_stop.tv_sec - iter_start.tv_sec) * 1000 + (iter_stop.tv_usec - iter_start.tv_usec)/1000.0) + 0.5);

   calcmuen(&mu, &en, psi, dpsi, psidd2, tmpxi, tmpyi, tmpzi, tmpxj, tmpyj, tmpzj);
   calcrms(rms, psi, tmpxi, tmpyi, tmpzi);
   psi2 = psi[Nx2][Ny2][Nz2] * psi[Nx2][Ny2][Nz2];
   fprintf(out, "After NPAS iter.:%8.4f %11.5f %11.5f %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
   fflush(out);
   if (Npasout != NULL) {
      sprintf(filename, "%s.txt", Npasout);
      file = fopen(filename, "w");
      outdenxyz(psi, file);
      fclose(file);

      sprintf(filename, "%s1d_x.txt", Npasout);
      file = fopen(filename, "w");
      outdenx(psi, *tmpyi, *tmpzi, file);
      fclose(file);

      sprintf(filename, "%s1d_y.txt", Npasout);
      file = fopen(filename, "w");
      outdeny(psi, *tmpxi, *tmpzi, file);
      fclose(file);

      sprintf(filename, "%s1d_z.txt", Npasout);
      file = fopen(filename, "w");
      outdenz(psi, *tmpxi, *tmpyi, file);
      fclose(file);
      /*
            sprintf(filename, "%s2d_xy.txt", Npasout);
            file = fopen(filename, "w");
            outdenxy(psi, *tmpzi, file);
            fclose(file);

            sprintf(filename, "%s2d_xz.txt", Npasout);
            file = fopen(filename, "w");
            outdenxz(psi, *tmpyi, file);
            fclose(file);

            sprintf(filename, "%s2d_yz.txt", Npasout);
            file = fopen(filename, "w");
            outdenyz(psi, *tmpxi, file);
            fclose(file);

            sprintf(filename, "%s3d_x0z.txt", Npasout);
            file = fopen(filename, "w");
            outpsi2xz(psi, file);
            fclose(file);

            sprintf(filename, "%s3d_xy0.txt", Npasout);
            file = fopen(filename, "w");
            outpsi2xy(psi, file);
            fclose(file);
      */
   }
   if (rmsout != NULL) {
      fprintf(filerms, "  After NPAS iter.:%12.5f %12.5f %12.5f %12.5f\n", rms[0], rms[1], rms[2], rms[3]);
      fflush(filerms);
   }

   gettimeofday(&iter_start, NULL);
   for (cnti = 0; cnti < Nrun; cnti ++) {
      calcnu(psi, psidd2);
      calclux(psi, cbeta);
      calcluy(psi, cbeta);
      calcluz(psi, cbeta);
      calcnorm(&norm, psi, tmpxi, tmpyi, tmpzi);
   }
   gettimeofday(&iter_stop, NULL);
   iter_time += (double) (((iter_stop.tv_sec - iter_start.tv_sec) * 1000 + (iter_stop.tv_usec - iter_start.tv_usec)/1000.0) + 0.5);

   calcmuen(&mu, &en, psi, dpsi, psidd2, tmpxi, tmpyi, tmpzi, tmpxj, tmpyj, tmpzj);
   calcrms(rms, psi, tmpxi, tmpyi, tmpzi);
   psi2 = psi[Nx2][Ny2][Nz2] * psi[Nx2][Ny2][Nz2];
   fprintf(out, "After NRUN iter.:%8.4f %11.5f %11.5f %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
   fflush(out);
   if (Nrunout != NULL) {

      sprintf(filename, "%s.txt", Nrunout);
      file = fopen(filename, "w");
      outdenxyz(psi, file);
      fclose(file);

      sprintf(filename, "%s1d_x.txt", Nrunout);
      file = fopen(filename, "w");
      outdenx(psi, *tmpyi, *tmpzi, file);
      fclose(file);

      sprintf(filename, "%s1d_y.txt", Nrunout);
      file = fopen(filename, "w");
      outdeny(psi, *tmpxi, *tmpzi, file);
      fclose(file);

      sprintf(filename, "%s1d_z.txt", Nrunout);
      file = fopen(filename, "w");
      outdenz(psi, *tmpxi, *tmpyi, file);
      fclose(file);
      /*
            sprintf(filename, "%s2d_xy.txt", Nrunout);
            file = fopen(filename, "w");
            outdenxy(psi, *tmpzi, file);
            fclose(file);

            sprintf(filename, "%s2d_xz.txt", Nrunout);
            file = fopen(filename, "w");
            outdenxz(psi, *tmpyi, file);
            fclose(file);

            sprintf(filename, "%s2d_yz.txt", Nrunout);
            file = fopen(filename, "w");
            outdenyz(psi, *tmpxi, file);
            fclose(file);

            sprintf(filename, "%s3d_x0z.txt", Nrunout);
            file = fopen(filename, "w");
            outpsi2xz(psi, file);
            fclose(file);

            sprintf(filename, "%s3d_xy0.txt", Nrunout);
            file = fopen(filename, "w");
            outpsi2xy(psi, file);
            fclose(file);
      */
   }
   if (rmsout != NULL) {
      fprintf(filerms, "  After NRUN iter.:%12.5f %12.5f %12.5f %12.5f\n", rms[0], rms[1], rms[2], rms[3]);
      fprintf(filerms, "                  --------------------------------------------------------\n");
   }
   if (rmsout != NULL) fclose(filerms);

   fprintf(out, "                  --------------------------------------------------------\n\n");

   free_double_vector(rms);

   free_double_vector(x);
   free_double_vector(y);
   free_double_vector(z);

   free_double_vector(x2);
   free_double_vector(y2);
   free_double_vector(z2);

   free_double_tensor(pot);
   free_double_tensor(potdd);
   free_double_tensor(psi);
   free_double_tensor(psidd2);

   free_double_tensor(dpsi);

   free_double_vector(calphax);
   free_double_vector(calphay);
   free_double_vector(calphaz);
   free_double_matrix(cbeta);
   free_double_vector(cgammax);
   free_double_vector(cgammay);
   free_double_vector(cgammaz);

   free_double_matrix(tmpxi);
   free_double_matrix(tmpyi);
   free_double_matrix(tmpzi);
   free_double_matrix(tmpxj);
   free_double_matrix(tmpyj);
   free_double_matrix(tmpzj);

   fftw_destroy_plan(plan_forward);
   fftw_destroy_plan(plan_backward);

   gettimeofday(&stop, NULL);
   wall_time = (double) (((stop.tv_sec - start.tv_sec) * 1000 + (stop.tv_usec - start.tv_usec)/1000.0) + 0.5);
   init_time = wall_time - iter_time;
   fprintf(out, " Initialization/allocation wall-clock time: %.3f seconds\n", init_time / 1000.);
   fprintf(out, " Calculation (iterations) wall-clock time: %.3f seconds\n", iter_time / 1000.);

   if(output != NULL) fclose(out);

   return (EXIT_SUCCESS);
}

/**
 *    Reading input parameters from the configuration file.
 */
void readpar(void) {
   char *cfg_tmp;

   if ((cfg_tmp = cfg_read("OPTION")) == NULL) {
      fprintf(stderr, "OPTION is not defined in the configuration file\n");
      exit(EXIT_FAILURE);
   }
   opt = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("NATOMS")) == NULL) {
      fprintf(stderr, "NATOMS is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Na = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("G0")) == NULL) {
      if ((cfg_tmp = cfg_read("AHO")) == NULL) {
         fprintf(stderr, "AHO is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      aho = atof(cfg_tmp);

      if ((cfg_tmp = cfg_read("AS")) == NULL) {
         fprintf(stderr, "AS is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      as = atof(cfg_tmp);

      g0 = 4. * pi * as * Na * BOHR_RADIUS / aho;
   } else {
      g0 = atof(cfg_tmp);
   }

   if ((cfg_tmp = cfg_read("GDD0")) == NULL) {
      if ((cfg_tmp = cfg_read("AHO")) == NULL) {
         fprintf(stderr, "AHO is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      aho = atof(cfg_tmp);

      if ((cfg_tmp = cfg_read("ADD")) == NULL) {
         fprintf(stderr, "ADD is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      add = atof(cfg_tmp);

      gd0 = 3. * add * Na * BOHR_RADIUS / aho;
   } else {
      gd0 = atof(cfg_tmp);
   }

   if ((cfg_tmp = cfg_read("NX")) == NULL) {
      fprintf(stderr, "NX is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Nx = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("NY")) == NULL) {
      fprintf(stderr, "NY is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Ny = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("NZ")) == NULL) {
      fprintf(stderr, "Nz is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Nz = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("DX")) == NULL) {
      fprintf(stderr, "DX is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   dx = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("DY")) == NULL) {
      fprintf(stderr, "DY is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   dy = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("DZ")) == NULL) {
      fprintf(stderr, "DZ is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   dz = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("DT")) == NULL) {
      fprintf(stderr, "DT is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   dt = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("GAMMA")) == NULL) {
      fprintf(stderr, "GAMMA is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   vgamma = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("NU")) == NULL) {
      fprintf(stderr, "NU is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   vnu = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("LAMBDA")) == NULL) {
      fprintf(stderr, "LAMBDA is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   vlambda = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("NSTP")) == NULL) {
      fprintf(stderr, "NSTP is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Nstp = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("NPAS")) == NULL) {
      fprintf(stderr, "NPAS is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Npas = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("NRUN")) == NULL) {
      fprintf(stderr, "NRUN is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Nrun = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("CUTOFF")) == NULL) {
      fprintf(stderr, "CUTOFF is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   cutoff = atof(cfg_tmp);

   output = cfg_read("OUTPUT");
   rmsout = cfg_read("RMSOUT");
   initout = cfg_read("INITOUT");
   Nstpout = cfg_read("NSTPOUT");
   Npasout = cfg_read("NPASOUT");
   Nrunout = cfg_read("NRUNOUT");

   if ((initout != NULL) || (Nstpout != NULL) || (Npasout != NULL) || (Nrunout != NULL)) {
      if ((cfg_tmp = cfg_read("OUTSTPX")) == NULL) {
         fprintf(stderr, "OUTSTPX is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      outstpx = atol(cfg_tmp);

      if ((cfg_tmp = cfg_read("OUTSTPY")) == NULL) {
         fprintf(stderr, "OUTSTPY is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      outstpy = atol(cfg_tmp);

      if ((cfg_tmp = cfg_read("OUTSTPZ")) == NULL) {
         fprintf(stderr, "OUTSTPZ is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      outstpz = atol(cfg_tmp);
   }

   return;
}

/**
 *    Initialization of the space mesh and the initial wave function.
 *    psi - array with the wave function values
 */
void initpsi(double ***psi) {
   long cnti, cntj, cntk;
   double cpsi;
   double tmp;

   cpsi = sqrt(pi * sqrt(pi / (vgamma * vnu * vlambda)));

   #pragma omp parallel for private(cnti)
   for (cnti = 0; cnti < Nx; cnti ++) {
      x[cnti] = (cnti - Nx2) * dx;
      x2[cnti] = x[cnti] * x[cnti];
   }

   #pragma omp parallel for private(cntj)
   for (cntj = 0; cntj < Ny; cntj ++) {
      y[cntj] = (cntj - Ny2) * dy;
      y2[cntj] = y[cntj] * y[cntj];
   }

   #pragma omp parallel for private(cntk)
   for (cntk = 0; cntk < Nz; cntk ++) {
      z[cntk] = (cntk - Nz2) * dz;
      z2[cntk] = z[cntk] * z[cntk];
   }

   #pragma omp parallel for private(cnti, cntj, cntk, tmp)
   for (cnti = 0; cnti < Nx; cnti ++) {
      for (cntj = 0; cntj < Ny; cntj ++) {
         for (cntk = 0; cntk < Nz; cntk ++) {
            tmp = exp(- 0.5 * (vgamma * x2[cnti] + vnu * y2[cntj] + vlambda * z2[cntk]));
            psi[cnti][cntj][cntk] = tmp / cpsi;
         }
      }
   }

   return;
}

/**
 *    Initialization of the potential.
 */
void initpot() {
   long cnti, cntj, cntk;
   double vnu2, vlambda2, vgamma2;

   vnu2 = vnu * vnu;
   vlambda2 = vlambda * vlambda;
   vgamma2 = vgamma * vgamma;

   #pragma omp parallel for private(cnti, cntj, cntk)
   for (cnti = 0; cnti < Nx; cnti ++) {
      for (cntj = 0; cntj < Ny; cntj ++) {
         for (cntk = 0; cntk < Nz; cntk ++) {
            pot[cnti][cntj][cntk] = 0.5 * par * (vgamma2 * x2[cnti] + vnu2 * y2[cntj] + vlambda2 * z2[cntk]);
         }
      }
   }

   return;
}

/**
 *    Crank-Nicolson scheme coefficients generation.
 */
void gencoef(void) {
   long cnti;

   Ax0 = 1. + dt / dx2 / (3. - par);
   Ay0 = 1. + dt / dy2 / (3. - par);
   Az0 = 1. + dt / dz2 / (3. - par);

   Ax0r = 1. - dt / dx2 / (3. - par);
   Ay0r = 1. - dt / dy2 / (3. - par);
   Az0r = 1. - dt / dz2 / (3. - par);

   Ax = - 0.5 * dt / dx2 / (3. - par);
   Ay = - 0.5 * dt / dy2 / (3. - par);
   Az = - 0.5 * dt / dz2 / (3. - par);

   calphax[Nx - 2] = 0.;
   cgammax[Nx - 2] = - 1. / Ax0;
   for (cnti = Nx - 2; cnti > 0; cnti --) {
      calphax[cnti - 1] = Ax * cgammax[cnti];
      cgammax[cnti - 1] = - 1. / (Ax0 + Ax * calphax[cnti - 1]);
   }

   calphay[Ny - 2] = 0.;
   cgammay[Ny - 2] = - 1. / Ay0;
   for (cnti = Ny - 2; cnti > 0; cnti --) {
      calphay[cnti - 1] = Ay * cgammay[cnti];
      cgammay[cnti - 1] = - 1. / (Ay0 + Ay * calphay[cnti - 1]);
   }

   calphaz[Nz - 2] = 0.;
   cgammaz[Nz - 2] = - 1. / Az0;
   for (cnti = Nz - 2; cnti > 0; cnti --) {
      calphaz[cnti - 1] = Az * cgammaz[cnti];
      cgammaz[cnti - 1] = - 1. / (Az0 + Az * calphaz[cnti - 1]);
   }

   return;
}

/**
 *    Initialization of the dipolar potential.
 *    kx  - array with the space mesh values in the x-direction in the K-space
 *    ky  - array with the space mesh values in the y-direction in the K-space
 *    kz  - array with the space mesh values in the z-direction in the K-space
 *    kx2 - array with the squared space mesh values in the x-direction in the
 *          K-space
 *    ky2 - array with the squared space mesh values in the y-direction in the
 *          K-space
 *    kz2 - array with the squared space mesh values in the z-direction in the
 *          K-space
 */
void initpotdd(double *kx, double *ky, double *kz, double *kx2, double *ky2, double *kz2) {
   long cnti, cntj, cntk;
   double dkx, dky, dkz, xk, tmp;

   dkx = 2. * pi / (Nx * dx);
   dky = 2. * pi / (Ny * dy);
   dkz = 2. * pi / (Nz * dz);

   for (cnti = 0; cnti < Nx2; cnti ++) kx[cnti] = cnti * dkx;
   for (cnti = 0; cnti < Nx2; cnti ++) kx[cnti + Nx2] = (cnti - Nx2) * dkx;
   for (cntj = 0; cntj < Ny2; cntj ++) ky[cntj] = cntj * dky;
   for (cntj = 0; cntj < Ny2; cntj ++) ky[cntj + Ny2] = (cntj - Ny2) * dky;
   for (cntk = 0; cntk < Nz2; cntk ++) kz[cntk] = cntk * dkz;
   for (cntk = 0; cntk < Nz2; cntk ++) kz[cntk + Nz2] = (cntk - Nz2) * dkz;

   for (cnti = 0; cnti < Nx; cnti ++) kx2[cnti] = kx[cnti] * kx[cnti];
   for (cntj = 0; cntj < Ny; cntj ++) ky2[cntj] = ky[cntj] * ky[cntj];
   for (cntk = 0; cntk < Nz; cntk ++) kz2[cntk] = kz[cntk] * kz[cntk];

   #pragma omp parallel for private(cnti, cntj, cntk, xk, tmp)
   for (cnti = 0; cnti < Nx; cnti ++) {
      for (cntj = 0; cntj < Ny; cntj ++) {
         for (cntk = 0; cntk < Nz; cntk ++) {
            xk = sqrt(kz2[cntk] + kx2[cnti] + ky2[cntj]);
            tmp = 1. + 3. * cos(xk * cutoff) / (xk * xk * cutoff * cutoff) - 3. * sin(xk * cutoff) / (xk * xk * xk * cutoff * cutoff * cutoff);
            potdd[cnti][cntj][cntk] = (4. * pi * (3. * kz2[cntk] / (kx2[cnti] + ky2[cntj] + kz2[cntk]) - 1.) / 3.) * tmp;
         }
      }
   }
   potdd[0][0][0] = 0.;

   return;
}

/**
 *    Calculation of the wave function norm and normalization.
 *    norm - wave function norm
 *    psi  - array with the wave function values
 *    tmpx - temporary array
 *    tmpy - temporary array
 *    tmpz - temporary array
 */
void calcnorm(double *norm, double ***psi, double **tmpx, double **tmpy, double **tmpz) {
   int threadid;
   long cnti, cntj, cntk;
   double tmp;

   #pragma omp parallel private(threadid, cnti, cntj, cntk)
   {
      threadid = omp_get_thread_num();

      #pragma omp for
      for (cnti = 0; cnti < Nx; cnti ++) {
         for (cntj = 0; cntj < Ny; cntj ++) {
            for (cntk = 0; cntk < Nz; cntk ++) {
               tmpz[threadid][cntk] = psi[cnti][cntj][cntk] * psi[cnti][cntj][cntk];
            }
            tmpy[threadid][cntj] = simpint(dz, tmpz[threadid], Nz);
         }
         (*tmpx)[cnti] = simpint(dy, tmpy[threadid], Ny);
      }
      #pragma omp barrier

      #pragma omp single
      {
         *norm = sqrt(simpint(dx, *tmpx, Nx));
         tmp = 1. / *norm;
      }

      #pragma omp for
      for (cnti = 0; cnti < Nx; cnti ++) {
         for (cntj = 0; cntj < Ny; cntj ++) {
            for (cntk = 0; cntk < Nz; cntk ++) {
               psi[cnti][cntj][cntk] *= tmp;
            }
         }
      }
   }

   return;
}

/**
 *    Calculation of the chemical potential and energy.
 *    mu     - chemical potential
 *    en     - energy
 *    psi    - array with the wave function values
 *    dpsi   - temporary array
 *    psidd2 - array with the squared wave function values
 *    tmpxi  - temporary array
 *    tmpyi  - temporary array
 *    tmpzi  - temporary array
 *    tmpxj  - temporary array
 *    tmpyj  - temporary array
 *    tmpzj  - temporary array
 */
void calcmuen(double *mu, double *en, double ***psi, double ***dpsi, double ***psidd2, double **tmpxi, double **tmpyi, double **tmpzi, double **tmpxj, double **tmpyj, double **tmpzj) {
   int threadid;
   long cnti, cntj, cntk;
   double psi2, psi2lin, psidd2lin, dpsi2;

   calcpsidd2(psi, psidd2);

   #pragma omp parallel private(threadid, cnti, cntj, cntk, psi2, psi2lin, psidd2lin, dpsi2)
   {
      threadid = omp_get_thread_num();

      #pragma omp for
      for (cntj = 0; cntj < Ny; cntj ++) {
         for (cntk = 0; cntk < Nz; cntk ++) {
            for (cnti = 0; cnti < Nx; cnti ++) {
               tmpxi[threadid][cnti] = psi[cnti][cntj][cntk];
            }
            diff(dx, tmpxi[threadid], tmpxj[threadid], Nx);
            for (cnti = 0; cnti < Nx; cnti ++) {
               dpsi[cnti][cntj][cntk] = tmpxj[threadid][cnti] * tmpxj[threadid][cnti];
            }
         }
      }

      #pragma omp for
      for (cnti = 0; cnti < Nx; cnti ++) {
         for (cntk = 0; cntk < Nz; cntk ++) {
            for (cntj = 0; cntj < Ny; cntj ++) {
               tmpyi[threadid][cntj] = psi[cnti][cntj][cntk];
            }
            diff(dy, tmpyi[threadid], tmpyj[threadid], Ny);
            for (cntj = 0; cntj < Ny; cntj ++) {
               dpsi[cnti][cntj][cntk] += tmpyj[threadid][cntj] * tmpyj[threadid][cntj];
            }
         }
      }

      #pragma omp for
      for (cnti = 0; cnti < Nx; cnti ++) {
         for (cntj = 0; cntj < Ny; cntj ++) {
            for (cntk = 0; cntk < Nz; cntk ++) {
               tmpzi[threadid][cntk] = psi[cnti][cntj][cntk];
            }
            diff(dz, tmpzi[threadid], tmpzj[threadid], Nz);
            for (cntk = 0; cntk < Nz; cntk ++) {
               dpsi[cnti][cntj][cntk] += tmpzj[threadid][cntk] * tmpzj[threadid][cntk];
            }
         }
      }
      #pragma omp barrier

      #pragma omp for
      for (cnti = 0; cnti < Nx; cnti ++) {
         for (cntj = 0; cntj < Ny; cntj ++) {
            for (cntk = 0; cntk < Nz; cntk ++) {
               psi2 = psi[cnti][cntj][cntk] * psi[cnti][cntj][cntk];
               psi2lin = psi2 * g;
               psidd2lin = psidd2[cnti][cntj][cntk] * gd;
               dpsi2 = dpsi[cnti][cntj][cntk] / (3. - par);
               tmpzi[threadid][cntk] = (pot[cnti][cntj][cntk] + psi2lin + psidd2lin) * psi2 + dpsi2;
               tmpzj[threadid][cntk] = (pot[cnti][cntj][cntk] + 0.5 * psi2lin + 0.5 * psidd2lin) * psi2 + dpsi2;
            }
            tmpyi[threadid][cntj] = simpint(dz, tmpzi[threadid], Nz);
            tmpyj[threadid][cntj] = simpint(dz, tmpzj[threadid], Nz);
         }
         (*tmpxi)[cnti] = simpint(dy, tmpyi[threadid], Ny);
         (*tmpxj)[cnti] = simpint(dy, tmpyj[threadid], Ny);
      }
   }

   *mu = simpint(dx, *tmpxi, Nx);
   *en = simpint(dx, *tmpxj, Nx);

   return;
}

/**
 *    Calculation of squared wave function values for dipole-dipole
 *    interaction.
 *    psi    - array with the wave function values
 *    psidd2 - array with the squared wave function values
 */
void calcpsidd2(double ***psi, double ***psidd2) {
   long cnti, cntj, cntk;
   fftw_complex *psidd2fft = (fftw_complex *) **psidd2;

   #pragma omp parallel for private(cnti, cntj, cntk)
   for (cnti = 0; cnti < Nx; cnti ++) {
      for (cntj = 0; cntj < Ny; cntj ++) {
         for (cntk = 0; cntk < Nz; cntk ++) {
            psidd2[cnti][cntj][cntk] = psi[cnti][cntj][cntk] * psi[cnti][cntj][cntk];
         }
      }
   }

   fftw_execute_dft_r2c(plan_forward, **psidd2, psidd2fft);

   #pragma omp parallel for private(cnti, cntj, cntk)
   for (cnti = 0; cnti < Nx; cnti ++) {
      for (cntj = 0; cntj < Ny; cntj ++) {
         for (cntk = 0; cntk < Nz2 + 1; cntk ++) {
            psidd2fft[cnti * Ny * (Nz2 + 1) + cntj * (Nz2 + 1) + cntk][0] *= potdd[cnti][cntj][cntk];
            psidd2fft[cnti * Ny * (Nz2 + 1) + cntj * (Nz2 + 1) + cntk][1] *= potdd[cnti][cntj][cntk];
         }
      }
   }

   fftw_execute_dft_c2r(plan_backward, psidd2fft, **psidd2);

   #pragma omp parallel for private(cnti, cntj, cntk)
   for (cnti = 0; cnti < Nx; cnti ++) {
      for (cntj = 0; cntj < Ny; cntj ++) {
         for (cntk = 0; cntk < Nz; cntk ++) {
            psidd2[cnti][cntj][cntk] /= Nx * Ny * Nz;
         }
      }
   }

   #pragma omp parallel for private(cntj, cntk)
   for (cntj = 0; cntj < Ny - 1; cntj ++) {
      for (cntk = 0; cntk < Nz - 1; cntk ++) {
         psidd2[Nx - 1][cntj][cntk] = psidd2[0][cntj][cntk];
      }
   }

   #pragma omp parallel for private(cnti, cntk)
   for (cnti = 0; cnti < Nx - 1; cnti ++) {
      for (cntk = 0; cntk < Nz - 1; cntk ++) {
         psidd2[cnti][Ny - 1][cntk] = psidd2[cnti][0][cntk];
      }
   }

   #pragma omp parallel for private(cnti, cntj)
   for (cnti = 0; cnti < Nx - 1; cnti ++) {
      for (cntj = 0; cntj < Ny - 1; cntj ++) {
         psidd2[cnti][cntj][Nz - 1] = psidd2[cnti][cntj][0];
      }
   }

   return;
}

/**
 *    Calculation of the root mean square radius.
 *    rms  - root mean square radius
 *    psi  - array with the wave function values
 *    tmpx - temporary array
 *    tmpy - temporary array
 *    tmpz - temporary array
 */
void calcrms(double *rms, double ***psi, double **tmpx, double **tmpy, double **tmpz) {
   int threadid;
   long cnti, cntj, cntk;
   double psi2;

   #pragma omp parallel private(threadid, cnti, cntj, cntk, psi2)
   {
      threadid = omp_get_thread_num();

      #pragma omp for
      for (cntj = 0; cntj < Ny; cntj ++) {
         for (cntk = 0; cntk < Nz; cntk ++) {
            for (cnti = 0; cnti < Nx; cnti ++) {
               psi2 = psi[cnti][cntj][cntk] * psi[cnti][cntj][cntk];
               tmpx[threadid][cnti] = x2[cnti] * psi2;
            }
            tmpz[threadid][cntk] = simpint(dx, tmpx[threadid], Nx);
         }
         (*tmpy)[cntj] = simpint(dz, tmpz[threadid], Nz);
      }
      #pragma omp barrier

      #pragma omp single
      rms[1] = sqrt(simpint(dy, *tmpy, Ny));

      #pragma omp for
      for (cnti = 0; cnti < Nx; cnti ++) {
         for (cntk = 0; cntk < Nz; cntk ++) {
            for (cntj = 0; cntj < Ny; cntj ++) {
               psi2 = psi[cnti][cntj][cntk] * psi[cnti][cntj][cntk];
               tmpy[threadid][cntj] = y2[cntj] * psi2;
            }
            tmpz[threadid][cntk] = simpint(dy, tmpy[threadid], Ny);
         }
         (*tmpx)[cnti] = simpint(dz, tmpz[threadid], Nz);
      }
      #pragma omp barrier

      #pragma omp single
      rms[2] = sqrt(simpint(dx, *tmpx, Nx));

      #pragma omp for
      for (cnti = 0; cnti < Nx; cnti ++) {
         for (cntj = 0; cntj < Ny; cntj ++) {
            for (cntk = 0; cntk < Nz; cntk ++) {
               psi2 = psi[cnti][cntj][cntk] * psi[cnti][cntj][cntk];
               tmpz[threadid][cntk] = z2[cntk] * psi2;
            }
            tmpy[threadid][cntj] = simpint(dz, tmpz[threadid], Nz);
         }
         (*tmpx)[cnti] = simpint(dy, tmpy[threadid], Ny);
      }
      #pragma omp barrier

      #pragma omp single
      rms[3] = sqrt(simpint(dx, *tmpx, Nx));
   }

   rms[0] = sqrt(rms[1] * rms[1] + rms[2] * rms[2] + rms[3] * rms[3]);

   return;
}

/**
 *    Time propagation with respect to H1 (part of the Hamiltonian without
 *    spatial derivatives).
 *    psi    - array with the wave function values
 *    psidd2 - array with the squared wave function values
 */
void calcnu(double ***psi, double ***psidd2) {
   long cnti, cntj, cntk;
   double psi2, psi2lin, psidd2lin, tmp;

   calcpsidd2(psi, psidd2);

   #pragma omp parallel for private(cnti, cntj, cntk, psi2, psi2lin, psidd2lin, tmp)
   for (cnti = 0; cnti < Nx; cnti ++) {
      for (cntj = 0; cntj < Ny; cntj ++) {
         for (cntk = 0; cntk < Nz; cntk ++) {
            psi2 = psi[cnti][cntj][cntk] * psi[cnti][cntj][cntk];
            psi2lin = psi2 * g;
            psidd2lin = psidd2[cnti][cntj][cntk] * gd;
            tmp = dt * (pot[cnti][cntj][cntk] + psi2lin + psidd2lin);
            psi[cnti][cntj][cntk] *= exp(- tmp);
         }
      }
   }

   return;
}

/**
 *    Time propagation with respect to H2 (x-part of the Laplacian).
 *    psi   - array with the wave function values
 *    cbeta - Crank-Nicolson scheme coefficients
 */
void calclux(double ***psi, double **cbeta) {
   int threadid;
   long cnti, cntj, cntk;
   double c;

   #pragma omp parallel private(threadid, cnti, cntj, cntk, c)
   {
      threadid = omp_get_thread_num();

      #pragma omp for
      for (cntj = 0; cntj < Ny; cntj ++) {
         for (cntk = 0; cntk < Nz; cntk ++) {
            cbeta[threadid][Nx - 2] = psi[Nx - 1][cntj][cntk];
            for (cnti = Nx - 2; cnti > 0; cnti --) {
               c = - Ax * psi[cnti + 1][cntj][cntk] + Ax0r * psi[cnti][cntj][cntk] - Ax * psi[cnti - 1][cntj][cntk];
               cbeta[threadid][cnti - 1] =  cgammax[cnti] * (Ax * cbeta[threadid][cnti] - c);
            }
            psi[0][cntj][cntk] = 0.;
            for (cnti = 0; cnti < Nx - 2; cnti ++) {
               psi[cnti + 1][cntj][cntk] = calphax[cnti] * psi[cnti][cntj][cntk] + cbeta[threadid][cnti];
            }
            psi[Nx - 1][cntj][cntk] = 0.;
         }
      }
   }

   return;
}

/**
 *    Time propagation with respect to H3 (y-part of the Laplacian).
 *    psi   - array with the wave function values
 *    cbeta - Crank-Nicolson scheme coefficients
 */
void calcluy(double ***psi, double **cbeta) {
   int threadid;
   long cnti, cntj, cntk;
   double c;

   #pragma omp parallel private(threadid, cnti, cntj, cntk, c)
   {
      threadid = omp_get_thread_num();

      #pragma omp for
      for (cnti = 0; cnti < Nx; cnti ++) {
         for (cntk = 0; cntk < Nz; cntk ++) {
            cbeta[threadid][Ny - 2] = psi[cnti][Ny - 1][cntk];
            for (cntj = Ny - 2; cntj > 0; cntj --) {
               c = - Ay * psi[cnti][cntj + 1][cntk] + Ay0r * psi[cnti][cntj][cntk] - Ay * psi[cnti][cntj - 1][cntk];
               cbeta[threadid][cntj - 1] =  cgammay[cntj] * (Ay * cbeta[threadid][cntj] - c);
            }
            psi[cnti][0][cntk] = 0.;
            for (cntj = 0; cntj < Ny - 2; cntj ++) {
               psi[cnti][cntj + 1][cntk] = calphay[cntj] * psi[cnti][cntj][cntk] + cbeta[threadid][cntj];
            }
            psi[cnti][Ny - 1][cntk] = 0.;
         }
      }
   }

   return;
}

/**
 *    Time propagation with respect to H4 (z-part of the Laplacian).
 *    psi   - array with the wave function values
 *    cbeta - Crank-Nicolson scheme coefficients
 */
void calcluz(double ***psi, double **cbeta) {
   int threadid;
   long cnti, cntj, cntk;
   double c;

   #pragma omp parallel private(threadid, cnti, cntj, cntk, c)
   {
      threadid = omp_get_thread_num();

      #pragma omp for
      for (cnti = 0; cnti < Nx; cnti ++) {
         for (cntj = 0; cntj < Ny; cntj ++) {
            cbeta[threadid][Nz - 2] = psi[cnti][cntj][Nz - 1];
            for (cntk = Nz - 2; cntk > 0; cntk --) {
               c = - Az * psi[cnti][cntj][cntk + 1] + Az0r * psi[cnti][cntj][cntk] - Az * psi[cnti][cntj][cntk - 1];
               cbeta[threadid][cntk - 1] =  cgammaz[cntk] * (Az * cbeta[threadid][cntk] - c);
            }
            psi[cnti][cntj][0] = 0.;
            for (cntk = 0; cntk < Nz - 2; cntk ++) {
               psi[cnti][cntj][cntk + 1] = calphaz[cntk] * psi[cnti][cntj][cntk] + cbeta[threadid][cntk];
            }
            psi[cnti][cntj][Nz - 1] = 0.;
         }
      }
   }

   return;
}

void outdenx(double ***psi, double *tmpy, double *tmpz, FILE *file) {
   long cnti, cntj, cntk;

   for (cnti = 0; cnti < Nx; cnti += outstpx) {
      for (cntj = 0; cntj < Ny; cntj ++) {
         for (cntk = 0; cntk < Nz; cntk ++) {
            tmpz[cntk] = psi[cnti][cntj][cntk] * psi[cnti][cntj][cntk];
         }
         tmpy[cntj] = simpint(dz, tmpz, Nz);
      }
      fprintf(file, "%8le %8le\n", x[cnti], simpint(dy, tmpy, Ny));
      fflush(file);
   }
}

void outdeny(double ***psi, double *tmpx, double *tmpz, FILE *file) {
   long cnti, cntj, cntk;

   for (cntj = 0; cntj < Ny; cntj += outstpy) {
      for (cnti = 0; cnti < Nx; cnti ++) {
         for (cntk = 0; cntk < Nz; cntk ++) {
            tmpz[cntk] = psi[cnti][cntj][cntk] * psi[cnti][cntj][cntk];
         }
         tmpx[cnti] = simpint(dz, tmpz, Nz);
      }
      fprintf(file, "%8le %8le\n", y[cntj], simpint(dx, tmpx, Nx));
      fflush(file);
   }
}

void outdenz(double ***psi, double *tmpx, double *tmpy, FILE *file) {
   long cnti, cntj, cntk;

   for (cntk = 0; cntk < Nz; cntk += outstpz) {
      for (cntj = 0; cntj < Ny; cntj ++) {
         for (cnti = 0; cnti < Nx; cnti ++) {
            tmpx[cnti] = psi[cnti][cntj][cntk] * psi[cnti][cntj][cntk];
         }
         tmpy[cntj] = simpint(dx, tmpx, Nx);
      }
      fprintf(file, "%8le %8le\n", z[cntk], simpint(dy, tmpy, Ny));
      fflush(file);
   }
}

void outdenxy(double ***psi, double *tmpz, FILE *file) {
   long cnti, cntj, cntk;

   for (cnti = 0; cnti < Nx; cnti += outstpx) {
      for (cntj = 0; cntj < Ny; cntj += outstpy) {
         for (cntk = 0; cntk < Nz; cntk ++) {
            tmpz[cntk] = psi[cnti][cntj][cntk] * psi[cnti][cntj][cntk];
         }
         fprintf(file, "%8le %8le %8le\n", x[cnti], y[cntj], simpint(dz, tmpz, Nz));
         fflush(file);
      }
   }

   return;
}

void outdenxz(double ***psi, double *tmpy, FILE *file) {
   long cnti, cntj, cntk;

   for (cnti = 0; cnti < Nx; cnti += outstpx) {
      for (cntk = 0; cntk < Nz; cntk += outstpz) {
         for (cntj = 0; cntj < Ny; cntj ++) {
            tmpy[cntj] = psi[cnti][cntj][cntk] * psi[cnti][cntj][cntk];
         }
         fprintf(file, "%8le %8le %8le\n", x[cnti], z[cntk], simpint(dy, tmpy, Ny));
         fflush(file);
      }
   }

   return;
}

void outdenxyz(double ***psi, FILE *file) {
   long cnti, cntj, cntk;

   for(cnti = 0; cnti < Nx; cnti += outstpx) {
      for(cntj = 0; cntj < Ny; cntj += outstpy) {
         for(cntk = 0; cntk < Nz; cntk += outstpz) {
            //fprintf(file, "%8le %8le %8le %8le\n", x[cnti], y[cntj], z[cntk], psi[cnti][cntj][cntk] * psi[cnti][cntj][cntk]);
            fprintf(file, "%8le\n", psi[cnti][cntj][cntk] * psi[cnti][cntj][cntk]);
            fflush(file);
         }
      }
   }

   return;
}

void outdenyz(double ***psi, double *tmpx, FILE *file) {
   long cnti, cntj, cntk;

   for (cntj = 0; cntj < Ny; cntj += outstpy) {
      for (cntk = 0; cntk < Nz; cntk += outstpz) {
         for (cnti = 0; cnti < Nx; cnti ++) {
            tmpx[cnti] = psi[cnti][cntj][cntk] * psi[cnti][cntj][cntk];
            fprintf(file, "%8le %8le %8le\n", y[cntj], z[cntk], simpint(dx, tmpx, Nx));
            fflush(file);
         }
      }
   }

   return;
}

void outpsi2xz(double ***psi, FILE *file) {
   long cnti, cntk;

   for (cnti = 0; cnti < Nx; cnti += outstpx) {
      for (cntk = 0; cntk < Nz; cntk += outstpz) {
         fprintf(file, "%8le %8le %8le\n", x[cnti], z[cntk], psi[cnti][Ny2][cntk] * psi[cnti][Ny2][cntk]);
      }
      fprintf(file, "\n");
      fflush(file);
   }

   return;
}

void outpsi2xy(double ***psi, FILE *file) {
   long cnti, cntj;

   for (cnti = 0; cnti < Nx; cnti += outstpx) {
      for (cntj = 0; cntj < Ny; cntj += outstpy) {
         fprintf(file, "%8le %8le %8le\n", x[cnti], y[cntj], psi[cnti][cntj][Nz2] * psi[cnti][cntj][Nz2]);
      }
      fprintf(file, "\n");
      fflush(file);
   }

   return;
}
