/**
 * DBEC-GP-OMP-CUDA-MPI programs are developed by:
 *
 * Vladimir Loncar, Antun Balaz
 * (Scientific Computing Laboratory, Institute of Physics Belgrade, Serbia)
 *
 * Srdjan Skrbic
 * (Department of Mathematics and Informatics, Faculty of Sciences, University of Novi Sad, Serbia)
 *
 * Paulsamy Muruganandam
 * (Bharathidasan University, Tamil Nadu, India)
 *
 * Luis E. Young-S, Sadhan K. Adhikari
 * (UNESP - Sao Paulo State University, Brazil)
 *
 *
 * Public use and modification of these codes are allowed provided that the
 * following papers are cited:
 * [1] V. Loncar et al., Comput. Phys. Commun. 209 (2016) 190.      
 * [2] V. Loncar et al., Comput. Phys. Commun. 200 (2016) 406.      
 * [3] R. Kishor Kumar et al., Comput. Phys. Commun. 195 (2015) 117.
 *
 * The authors would be grateful for all information and/or comments
 * regarding the use of the programs.
 */

#include "real1dX-th.h"

int main(int argc, char **argv) {
   FILE *out;
   FILE *file;
   FILE *dyna;
   char filename[MAX_FILENAME_SIZE];
   int nthreads;
   long cnti;
   double norm, mu, en;
   double *rms;
   double complex *psi;
   double *psidd2;
   double complex *cbeta;
   double *dpsix;
   double *tmpxi, *tmpxj, *tmpxk, *tmpxl, *tmpxm, *tmpxn;
   fftw_complex *psidd2fft;
   double psi2;

   pi = 3.14159265358979;

   struct timeval start, stop, iter_start, iter_stop;
   double wall_time, init_time, iter_time;
   iter_time = 0.;
   gettimeofday(&start, NULL);

   if((argc != 3) || (strcmp(*(argv + 1), "-i") != 0)) {
      fprintf(stderr, "Usage: %s -i <input parameter file> \n", *argv);
      exit(EXIT_FAILURE);
   }

   if(! cfg_init(argv[2])) {
      fprintf(stderr, "Wrong input parameter file.\n");
      exit(EXIT_FAILURE);
   }

   readpar();

   #pragma omp parallel
   #pragma omp master
   nthreads = omp_get_num_threads();

   rms = alloc_double_vector(RMS_ARRAY_SIZE);

   x = alloc_double_vector(Nx);

   x2 = alloc_double_vector(Nx);

   pot = alloc_double_vector(Nx);
   potdd = alloc_double_vector(Nx);

   psi = alloc_complex_vector(Nx);
   psidd2 = alloc_double_vector((Nx / 2 + 1) * 2);
   psidd2fft = (fftw_complex *) psidd2;

   dpsix = alloc_double_vector(Nx);

   calphax = alloc_complex_vector(Nx - 1);
   cbeta =  alloc_complex_vector(Nx - 1);
   cgammax = alloc_complex_vector(Nx - 1);

   tmpxi = alloc_double_vector(Nx);
   tmpxj = alloc_double_vector(Nx);
   tmpxk = alloc_double_vector(Nx);
   tmpxl = alloc_double_vector(Nx);
   tmpxm = alloc_double_vector(Nx);
   tmpxn = alloc_double_vector(Nx);

   if(output != NULL) {
      sprintf(filename, "%s.txt", output);
      out = fopen(filename, "w");
   }
   else out = stdout;

   if(opt == 2) par = 2.;
   else par = 1.;

   drho2 = drho * drho;
   g_1d = g0 / ((2. * pi * drho2));
   gd_1d = gd0 / ((2. * pi* drho2));

   Nx2 = Nx / 2;
   dx2 = dx * dx;

   fprintf(out, " Real time propagation 1dX,   OPTION = %d, OMP_NUM_THREADS = %d\n\n", opt, nthreads);
   fprintf(out, "  Number of Atoms N = %li, Unit of length AHO = %.8f m\n", Na,aho);
   fprintf(out, "  Scattering length a = %.2f*a0, Dipolar ADD = %.2f*a0\n", as, add);
   fprintf(out, "  Nonlinearity G_3D = %.4f, Strength of DDI GD_3D = %.5f\n", g0, gd0);
   fprintf(out, "  Nonlinearity G_1D = %.4f, Strength of DDI GD_1D = %.5f\n", g_1d, gd_1d);
   fprintf(out, "  Parameter of trap: GAMMA = %.2f\n", vgamma);
   fprintf(out, "  Radial Gaussian Width = %.2f\n\n", drho);
   fprintf(out, " # Space Stp:  N = %li\n", Nx);
   fprintf(out, "  Space Step: DX = %.6f\n", dx);
   fprintf(out, " # Time Stp : NSTP = %li, NPAS = %li, NRUN = %li\n", Nstp, Npas, Nrun);
   fprintf(out, "   Time Step:   DT = %.6f\n\n",  dt);
   fprintf(out, " * Change for dynamics: GPAR = %.3f, GDPAR = %.3f *\n\n", gpar, gdpar);
   fprintf(out, "                  --------------------------------------------------------\n");
   fprintf(out, "                    Norm      Chem        Ener/N      <x>      |Psi(0)|^2\n");
   fprintf(out, "                  --------------------------------------------------------\n");
   fflush(out);

   if(Nstp == 0) {
     g = par * g_1d;
     gd = par * gd0;
   }
   else {
     g = 0.;
     gd = 0.;
   }

   fftw_init_threads();
   fftw_plan_with_nthreads(nthreads);
   initpsi(psi);
   initpot();
   gencoef();
   initpotdd(tmpxi, tmpxj, tmpxk, tmpxl, tmpxm, tmpxn);

   plan_forward = fftw_plan_dft_r2c_1d(Nx, psidd2, psidd2fft, FFT_FLAG);
   plan_backward = fftw_plan_dft_c2r_1d(Nx, psidd2fft, psidd2, FFT_FLAG);


   calcnorm(&norm, psi, tmpxi);
   calcmuen(&mu, &en, psi, dpsix, psidd2, tmpxi, tmpxj);
   calcrms(rms, psi, tmpxi);
   psi2 = cabs(psi[Nx2]);
   psi2 *= psi2;
   fprintf(out, "Initial : %15.4f %11.5f %11.5f %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
   fflush(out);
   if(initout != NULL) {
      sprintf(filename, "%s.txt", initout);
      file = fopen(filename, "w");
      outpsi2x(psi, file);
      fclose(file);
   }

   if(Nstp != 0) {
      double g_stp = par * g_1d / (double) Nstp;
      double gd_stp = par * gd0 / (double) Nstp;
      gettimeofday(&iter_start, NULL);
      for(cnti = 0; cnti < Nstp; cnti ++) {
         g += g_stp;
         gd += gd_stp;
         calcnu(psi, psidd2);
         calclux(psi, cbeta);
      }
      gettimeofday(&iter_stop, NULL);
      iter_time += (double) (((iter_stop.tv_sec - iter_start.tv_sec) * 1000 + (iter_stop.tv_usec - iter_start.tv_usec)/1000.0) + 0.5);

      calcnorm(&norm, psi, tmpxi);
      calcmuen(&mu, &en, psi, dpsix, psidd2, tmpxi, tmpxj);
      calcrms(rms, psi, tmpxi);
      psi2 = cabs(psi[Nx2]);
      psi2 *= psi2;
      fprintf(out, "After NSTP iter.:%8.4f %11.5f %11.5f %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
      fflush(out);
      if(Nstpout != NULL) {
         sprintf(filename, "%s.txt", Nstpout);
         file = fopen(filename, "w");
         outpsi2x(psi, file);
         fclose(file);
      }
   }

   if(dynaout != NULL) {
      sprintf(filename, "%s.txt", dynaout);
      dyna = fopen(filename, "w");
   }
   else dyna = NULL;

   gettimeofday(&iter_start, NULL);
   for(cnti = 0; cnti < Npas; cnti ++) {
      calcnu(psi, psidd2);
      calclux(psi, cbeta);
      if((dynaout != NULL) && (cnti % outstpt == 0)) {
         calcrms(rms, psi, tmpxi);
         fprintf(dyna, "%8le   %8le\n", cnti * dt * par, *rms);
         fflush(dyna);
      }
   }
   gettimeofday(&iter_stop, NULL);
   iter_time += (double) (((iter_stop.tv_sec - iter_start.tv_sec) * 1000 + (iter_stop.tv_usec - iter_start.tv_usec)/1000.0) + 0.5);

   calcnorm(&norm, psi, tmpxi);
   calcmuen(&mu, &en, psi, dpsix, psidd2, tmpxi, tmpxj);
   calcrms(rms, psi, tmpxi);
   psi2 = cabs(psi[Nx2]);
   psi2 *= psi2;
   fprintf(out, "After NPAS iter.:%8.4f %11.5f %11.5f %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
   fflush(out);
   if(Npasout != NULL) {
      sprintf(filename, "%s.txt", Npasout);
      file = fopen(filename, "w");
      outpsi2x(psi, file);
      fclose(file);
   }

   g = gpar * g;
   gd = gdpar * gd;

   gettimeofday(&iter_start, NULL);
   for(cnti = 0; cnti < Nrun; cnti ++) {
      calcnu(psi, psidd2);
      calclux(psi, cbeta);
      if((dynaout != NULL) && (cnti % outstpt == 0)) {
         calcrms(rms, psi, tmpxi);
         fprintf(dyna, "%8le   %8le\n", (cnti * dt  + dt * Npas) * par, *rms);
         fflush(dyna);
      }

   }
   gettimeofday(&iter_stop, NULL);
   iter_time += (double) (((iter_stop.tv_sec - iter_start.tv_sec) * 1000 + (iter_stop.tv_usec - iter_start.tv_usec)/1000.0) + 0.5);

   if(dynaout != NULL) fclose(dyna);
   calcnorm(&norm, psi, tmpxi);
   calcmuen(&mu, &en, psi, dpsix, psidd2, tmpxi, tmpxj);
   calcrms(rms, psi, tmpxi);
   psi2 = cabs(psi[Nx2]);
   psi2 *= psi2;
   fprintf(out, "After NRUN iter.:%8.4f %11.5f %11.5f %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
   fflush(out);
   if(Nrunout != NULL) {
      sprintf(filename, "%s.txt", Nrunout);
      file = fopen(filename, "w");
      outpsi2x(psi, file);
      fclose(file);
   }
   fprintf(out, "                  --------------------------------------------------------\n\n");

   free_double_vector(rms);

   free_double_vector(x);

   free_double_vector(x2);

   free_double_vector(pot);
   free_double_vector(potdd);
   free_complex_vector(psi);
   free_double_vector(psidd2);

   free_double_vector(dpsix);

   free_complex_vector(calphax);
   free_complex_vector(cbeta);
   free_complex_vector(cgammax);

   free_double_vector(tmpxi);
   free_double_vector(tmpxj);
   free_double_vector(tmpxk);
   free_double_vector(tmpxl);
   free_double_vector(tmpxm);
   free_double_vector(tmpxn);

   fftw_destroy_plan(plan_forward);
   fftw_destroy_plan(plan_backward);

   gettimeofday(&stop, NULL);
   wall_time = (double) (((stop.tv_sec - start.tv_sec) * 1000 + (stop.tv_usec - start.tv_usec)/1000.0) + 0.5);
   init_time = wall_time - iter_time;
   fprintf(out, " Initialization/allocation wall-clock time: %.3f seconds\n", init_time / 1000.);
   fprintf(out, " Calculation (iterations) wall-clock time: %.3f seconds\n", iter_time / 1000.);

   if(output != NULL) fclose(out);

   return(EXIT_SUCCESS);
}

/**
 *    Reading input parameters from the configuration file.
 */
void readpar(void) {
   char *cfg_tmp;

   if((cfg_tmp = cfg_read("OPTION")) == NULL) {
      fprintf(stderr, "OPTION is not defined in the configuration file\n");
      exit(EXIT_FAILURE);
   }
   opt = atol(cfg_tmp);

   if((cfg_tmp = cfg_read("NATOMS")) == NULL) {
      fprintf(stderr, "NATOMS is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Na = atol(cfg_tmp);

   if((cfg_tmp = cfg_read("G0")) == NULL) {
      if((cfg_tmp = cfg_read("AHO")) == NULL) {
         fprintf(stderr, "AHO is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      aho = atof(cfg_tmp);

      if((cfg_tmp = cfg_read("AS")) == NULL) {
         fprintf(stderr, "AS is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      as = atof(cfg_tmp);

      g0 = 4. * pi * as * Na * BOHR_RADIUS / aho;
   } else {
      g0 = atof(cfg_tmp);
   }

   if((cfg_tmp = cfg_read("GPAR")) == NULL) {
      fprintf(stderr, "GPAR is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   gpar = atof(cfg_tmp);

   if((cfg_tmp = cfg_read("GDD0")) == NULL) {
      if((cfg_tmp = cfg_read("AHO")) == NULL) {
         fprintf(stderr, "AHO is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      aho = atof(cfg_tmp);

      if((cfg_tmp = cfg_read("ADD")) == NULL) {
         fprintf(stderr, "ADD is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      add = atof(cfg_tmp);

      gd0 = 3. * add * Na * BOHR_RADIUS / aho;
   } else {
      gd0 = atof(cfg_tmp);
   }

   if((cfg_tmp = cfg_read("GDDPAR")) == NULL) {
      fprintf(stderr, "GDDPAR is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   gdpar = atof(cfg_tmp);

   if((cfg_tmp = cfg_read("NX")) == NULL) {
      fprintf(stderr, "NX is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Nx = atol(cfg_tmp);

   if((cfg_tmp = cfg_read("DX")) == NULL) {
      fprintf(stderr, "DX is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   dx = atof(cfg_tmp);

   if((cfg_tmp = cfg_read("DT")) == NULL) {
      fprintf(stderr, "DT is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   dt = atof(cfg_tmp);

   if((cfg_tmp = cfg_read("GAMMA")) == NULL) {
      fprintf(stderr, "GAMMA is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   vgamma = atof(cfg_tmp);

   if((cfg_tmp = cfg_read("DRHO")) == NULL) {
      fprintf(stderr, "DRHO is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   drho = atof(cfg_tmp);

   if((cfg_tmp = cfg_read("NSTP")) == NULL) {
      fprintf(stderr, "NSTP is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Nstp = atol(cfg_tmp);

   if((cfg_tmp = cfg_read("NPAS")) == NULL) {
      fprintf(stderr, "NPAS is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Npas = atol(cfg_tmp);

   if((cfg_tmp = cfg_read("NRUN")) == NULL) {
      fprintf(stderr, "NRUN is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Nrun = atol(cfg_tmp);

   output = cfg_read("OUTPUT");
   dynaout = cfg_read("DYNAOUT");
   initout = cfg_read("INITOUT");
   Nstpout = cfg_read("NSTPOUT");
   Npasout = cfg_read("NPASOUT");
   Nrunout = cfg_read("NRUNOUT");

   if((initout != NULL) || (Nstpout != NULL) || (Npasout != NULL) || (Nrunout != NULL)) {
      if((cfg_tmp = cfg_read("OUTSTPX")) == NULL) {
         fprintf(stderr, "OUTSTPX is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      outstpx = atol(cfg_tmp);
   }

   if(dynaout != NULL) {
      if((cfg_tmp = cfg_read("OUTSTPT")) == NULL) {
         fprintf(stderr, "OUTSTPT is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      outstpt = atol(cfg_tmp);
   }

   return;
}

/**
 *    Initialization of the space mesh and the initial wave function.
 *    psi - array with the wave function values
 */
void initpsi(double complex *psi) {
   long cnti;
   double cpsi;
   double tmp;
   FILE *file;

   cpsi = sqrt( sqrt(pi  / vgamma));

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nx; cnti ++) {
      x[cnti] = (cnti - Nx2) * dx;
      x2[cnti] = x[cnti] * x[cnti];
   }

   if(Nstp != 0) {
      #pragma omp parallel for private(cnti, tmp)
      for(cnti = 0; cnti < Nx; cnti ++) {
         tmp = exp(- 0.5 * vgamma * x2[cnti]);
         psi[cnti] = tmp / cpsi;
      }
   }

   if(Nstp == 0) {
      if((file = fopen("imag1dX-den.txt", "r")) == NULL) {       /* open a text file for reading */
         printf("Run the program using the input file to read. i.g.: imag1dX-den.txt\n");
         exit(1);					/*couldn't open the requested file!*/
      }
      file = fopen("imag1dX-den.txt", "r");
      for(cnti = 0; cnti < Nx; cnti ++) {
         if(fscanf(file, "%lf %lf\n", &x[cnti], &tmp)) {
            psi[cnti] = sqrt(tmp);
         }
      }
      fclose(file);
   }

   return;
}

/**
 *    Initialization of the potential.
 */
void initpot() {
   long cnti;
   double vgamma2;

   vgamma2 = vgamma * vgamma;

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nx; cnti ++) {
      pot[cnti] = 0.5 * par * (vgamma2 * x2[cnti]);
   }

   return;
}

/**
 *    Crank-Nicolson scheme coefficients generation.
 */
void gencoef(void) {
   long cnti;

   Ax0 = 1. + I * dt / dx2 / (3. - par);

   Ax0r = 1. - I * dt / dx2 / (3. - par);

   Ax = - 0.5 * I * dt / dx2 / (3. - par);

   calphax[Nx - 2] = 0.;
   cgammax[Nx - 2] = - 1. / Ax0;
   for (cnti = Nx - 2; cnti > 0; cnti --) {
      calphax[cnti - 1] = Ax * cgammax[cnti];
      cgammax[cnti - 1] = - 1. / (Ax0 + Ax * calphax[cnti - 1]);
   }

   return;
}

/**
 *    Initialization of the dipolar potential.
 *    kx  - array with the space mesh values in the x-direction in the K-space
 *    ky  - array with the space mesh values in the y-direction in the K-space
 *    kx2 - array with the squared space mesh values in the x-direction in the
 *          K-space
 *    ky2 - array with the squared space mesh values in the y-direction in the
 *          K-space
 */
void initpotdd(double *kx, double *kx2, double *xg, double *dxg, double *ky, double *dky) {
   long cnti, cntj;
   double ky2;
   double tau;

   gauleg(-1., 1., xg, dxg, Nx);

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nx; cnti ++) {
      ky[cnti] = xg[cnti] / (1. - xg[cnti] * xg[cnti]);
      kx[cnti] = ky[cnti];
      dky[cnti] = dxg[cnti] * (1. + xg[cnti] * xg[cnti]) /
      ((1. - xg[cnti] * xg[cnti]) * (1. - xg[cnti] * xg[cnti]));
   }

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nx; cnti ++) kx2[cnti] = kx[cnti] * kx[cnti];

   #pragma omp parallel for private(cnti, cntj, ky2, tau)
   for(cnti = 0; cnti < Nx; cnti ++) {
      potdd[cnti] = 0.;
      for(cntj = 0; cntj < Nx; cntj ++) {
         ky2 = ky[cntj] * ky[cntj];
         tau = sqrt((kx2[cntj] + ky2) * drho2 / 2.);
         if(tau < 26.) {
            potdd[cnti] += dky[cntj] * sqrt(drho2 / 2.) * exp(- ky2 * drho2 / 2.)
	          * (2. - 3. * sqrt(pi) * exp(tau * tau) * tau * erfc(tau));
         } else {
            potdd[cnti] += dky[cntj] * sqrt(drho2 / 2.) * exp(- ky2 * drho2 / 2.)
	          * (-1. + .5 / (tau * tau));
         }
      }
      potdd[cnti] *= 2. / (3. * sqrt(pi));
   }

   return;
}

/**
 *    Calculation of the wave function norm and normalization.
 *    norm - wave function norm
 *    psi  - array with the wave function values
 *    tmpx - temporary array
 *    tmpy - temporary array
 */
void calcnorm(double *norm, double complex *psi, double *tmpx) {
   long cnti;
   //double tmp;

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nx; cnti ++) {
      tmpx[cnti] = cabs(psi[cnti]);
      tmpx[cnti] *= tmpx[cnti];
   }

   *norm = sqrt(simpint_th(dx, tmpx, Nx));
   //tmp = 1. / *norm;

   //for(cnti = 0; cnti < Nx; cnti ++) {
   //   psi[cnti] *= tmp;
   //}

   return;
}

/**
 *    Calculation of the chemical potential and energy.
 *    mu     - chemical potential
 *    en     - energy
 *    psi    - array with the wave function values
 *    dpsix  - temporary array
 *    dpsiy  - temporary array
 *    psidd2 - array with the squared wave function values
 *    tmpxi  - temporary array
 *    tmpyi  - temporary array
 *    tmpzi  - temporary array
 *    tmpxj  - temporary array
 *    tmpyj  - temporary array
 */
void calcmuen(double *mu, double *en, double complex *psi, double *dpsix, double *psidd2, double *tmpxi, double *tmpxj) {
   long cnti;
   double psi2, psi2lin, psidd2lin, dpsi2;

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nx; cnti ++) {
      tmpxi[cnti] = cabs(psi[cnti]);
   }

   diff_th(dx, tmpxi, dpsix, Nx);

   calcpsidd2(psi, psidd2);

   #pragma omp parallel for private(cnti, psi2, psi2lin, psidd2lin, dpsi2)
   for(cnti = 0; cnti < Nx; cnti ++) {
      psi2 = cabs(psi[cnti]);
      psi2 *= psi2;
      psi2lin = psi2 * g;
      psidd2lin = psidd2[cnti] * gd;
      dpsi2 = dpsix[cnti] * dpsix[cnti] / (3. - par);
      tmpxi[cnti] = (pot[cnti] + psi2lin + psidd2lin) * psi2 + dpsi2;
      tmpxj[cnti] = (pot[cnti] + 0.5 * psi2lin + 0.5 * psidd2lin) * psi2 + dpsi2;
   }

   *mu = simpint_th(dx, tmpxi, Nx);
   *en = simpint_th(dx, tmpxj, Nx);

   return;
}

/**
 *    Calculation of squared wave function values for dipole-dipole
 *    interaction.
 *    psi    - array with the wave function values
 *    psidd2 - array with the squared wave function values
 */
void calcpsidd2(double complex *psi, double *psidd2) {
   long cnti;
   double tmp;
   fftw_complex *psidd2fft = (fftw_complex *) psidd2;

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nx; cnti ++) {
      tmp = cabs(psi[cnti]);
      psidd2[cnti] = tmp * tmp;
   }

   fftw_execute_dft_r2c(plan_forward, psidd2, psidd2fft);

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nx2 + 1; cnti ++) {
      psidd2fft[cnti][0] *= potdd[cnti];
      psidd2fft[cnti][1] *= potdd[cnti];
   }

   fftw_execute_dft_c2r(plan_backward, psidd2fft, psidd2);

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nx; cnti ++) {
      psidd2[cnti] /= Nx;
   }

   psidd2[Nx - 1] = psidd2[0];

   return;
}

/**
 *    Calculation of the root mean square radius.
 *    rms  - root mean square radius
 *    psi  - array with the wave function values
 *    tmpx - temporary array
 *    tmpy - temporary array
 */
void calcrms(double *rms, double complex *psi, double *tmpx) {
   long cnti;
   double psi2;

   #pragma omp parallel for private(cnti)
   for(cnti = 0; cnti < Nx; cnti ++) {
      psi2 = cabs(psi[cnti]);
      psi2 *= psi2;
      tmpx[cnti] = x2[cnti] * psi2;
   }

   *rms = sqrt(simpint_th(dx, tmpx, Nx));

   return;
}

/**
 *    Time propagation with respect to H1 (part of the Hamiltonian without
 *    spatial derivatives).
 *    psi    - array with the wave function values
 *    psidd2 - array with the squared wave function values
 */
void calcnu(double complex *psi, double *psidd2) {
   long cnti;
   double psi2, psi2lin, psidd2lin, tmp;

   calcpsidd2(psi, psidd2);

   #pragma omp parallel for private(cnti, psi2, psi2lin, psidd2lin, tmp)
   for(cnti = 0; cnti < Nx; cnti ++) {
      psi2 = cabs(psi[cnti]);
      psi2 *= psi2;
      psi2lin = psi2 * g;
      psidd2lin = psidd2[cnti] * gd;
      tmp = dt * (pot[cnti] + psi2lin + psidd2lin);
      psi[cnti] *= cexp(- I * tmp);
   }

   return;
}

/**
 *    Time propagation with respect to H2 (x-part of the Laplacian).
 *    psi   - array with the wave function values
 *    cbeta - Crank-Nicolson scheme coefficients
 */
void calclux(double complex *psi, double complex *cbeta) {
   long cnti;
   double complex c;

   cbeta[Nx - 2] = psi[Nx - 1];
   for (cnti = Nx - 2; cnti > 0; cnti --) {
      c = - Ax * psi[cnti + 1] + Ax0r * psi[cnti] - Ax * psi[cnti - 1];
      cbeta[cnti - 1] =  cgammax[cnti] * (Ax * cbeta[cnti] - c);
   }
   psi[0] = 0.;
   for (cnti = 0; cnti < Nx - 2; cnti ++) {
      psi[cnti + 1] = calphax[cnti] * psi[cnti] + cbeta[cnti];
   }
   psi[Nx - 1] = 0.;

   return;
}

void outpsi2x(double complex *psi, FILE *file) {
   long cnti;

   for(cnti = 0; cnti < Nx; cnti += outstpx) {
      fprintf(file, "%8le %8le\n", x[cnti], cabs(psi[cnti]) * cabs(psi[cnti]));
   }
   fflush(file);

   return;
}
