/**
 * DBEC-GP-OMP-CUDA-MPI programs are developed by:
 *
 * Vladimir Loncar, Antun Balaz
 * (Scientific Computing Laboratory, Institute of Physics Belgrade, Serbia)
 *
 * Srdjan Skrbic
 * (Department of Mathematics and Informatics, Faculty of Sciences, University of Novi Sad, Serbia)
 *
 * Paulsamy Muruganandam
 * (Bharathidasan University, Tamil Nadu, India)
 *
 * Luis E. Young-S, Sadhan K. Adhikari
 * (UNESP - Sao Paulo State University, Brazil)
 *
 *
 * Public use and modification of these codes are allowed provided that the
 * following papers are cited:
 * [1] V. Loncar et al., Comput. Phys. Commun. 209 (2016) 190.      
 * [2] V. Loncar et al., Comput. Phys. Commun. 200 (2016) 406.      
 * [3] R. Kishor Kumar et al., Comput. Phys. Commun. 195 (2015) 117.
 *
 * The authors would be grateful for all information and/or comments
 * regarding the use of the programs.
 */

#include "real2dXY-th.h"

int main(int argc, char **argv) {
   FILE *out;
   FILE *file;
   FILE *filerms;
   FILE *dyna;
   char filename[MAX_FILENAME_SIZE];
   int nthreads;
   long cnti;
   double norm, mu, en;
   double *rms;
   double complex **cbeta;
   double **dpsi;
   double **tmpxi, **tmpyi, **tmpxj, **tmpyj;
   double psi2;
   fftw_complex *psidd2fft;

   pi = 3.14159265358979;

   struct timeval start, stop, iter_start, iter_stop;
   double wall_time, init_time, iter_time;
   iter_time = 0.;
   gettimeofday(&start, NULL);

   file = NULL;

   if ((argc != 3) || (strcmp(*(argv + 1), "-i") != 0)) {
      fprintf(stderr, "Usage: %s -i <input parameter file> \n", *argv);
      exit(EXIT_FAILURE);
   }

   if (! cfg_init(argv[2])) {
      fprintf(stderr, "Wrong input parameter file.\n");
      exit(EXIT_FAILURE);
   }

   readpar();

   #pragma omp parallel
   #pragma omp master
   nthreads = omp_get_num_threads();

   rms = alloc_double_vector(RMS_ARRAY_SIZE);

   x = alloc_double_vector(Nx);
   y = alloc_double_vector(Ny);

   x2 = alloc_double_vector(Nx);
   y2 = alloc_double_vector(Ny);

   pot = alloc_double_matrix(Nx, Ny);
   potdd = alloc_double_matrix(Nx, Ny);
   psi = alloc_complex_matrix(Nx, Ny);
   psidd2 = alloc_double_matrix(Nx, (Ny / 2 + 1) * 2);
   psidd2fft = (fftw_complex *) *psidd2;

   dpsi = alloc_double_matrix(Nx, Ny);

   calphax = alloc_complex_vector(Nx - 1);
   calphay = alloc_complex_vector(Ny - 1);
   cbeta =  alloc_complex_matrix(nthreads, MAX(Nx, Ny) - 1);
   cgammax = alloc_complex_vector(Nx - 1);
   cgammay = alloc_complex_vector(Ny - 1);

   tmpxi = alloc_double_matrix(nthreads, Nx);
   tmpyi = alloc_double_matrix(nthreads, Ny);
   tmpxj = alloc_double_matrix(nthreads, Nx);
   tmpyj = alloc_double_matrix(nthreads, Ny);

   if (output != NULL) {
      sprintf(filename, "%s.txt", output);
      out = fopen(filename, "w");
   } else out = stdout;

   if (rmsout != NULL) {
      sprintf(filename, "%s.txt", rmsout);
      filerms = fopen(filename, "w");
   } else filerms = NULL;

   if (opt == 2) par = 2.;
   else par = 1.;

   g_2d = g0 / (sqrt(2. * pi)  * d_z);
   gd_2d = gd0 / (sqrt(2. * pi) * d_z);

   Nx2 = Nx / 2; Ny2 = Ny / 2;
   dx2 = dx * dx; dy2 = dy * dy;

   fprintf(out, " Real time propagation 2dXY,   OPTION = %d, OMP_NUM_THREADS = %d\n\n", opt, nthreads);
   fprintf(out, "  Number of Atoms N = %li, Unit of length AH0 = %.8f m\n", Na, aho);
   fprintf(out, "  Scattering length a = %.2f*a0, Dipolar ADD = %.2f*a0\n", as, add);
   fprintf(out, "  Nonlinearity G_3D = %.4f, Strength of DDI GD_3D = %.5f\n", g0, gd0);
   fprintf(out, "  Nonlinearity G_2D = %.4f, Strength of DDI GD_2D = %.5f\n", g_2d, gd_2d);
   fprintf(out, "  Parameters of trap: GAMMA = %.2f, NU = %.2f\n", vgamma, vnu);
   fprintf(out, "  Axial Gaussian Width = %.2f\n\n", d_z);
   fprintf(out, " # Space Stp: NX = %li, NY = %li\n", Nx, Ny);
   fprintf(out, "  Space Step: DX = %.6f, DY = %.6f\n", dx, dy);
   fprintf(out, " # Time Stp : NSTP = %li, NPAS = %li, NRUN = %li\n", Nstp, Npas, Nrun);
   fprintf(out, "   Time Step:   DT = %.6f\n\n",  dt);
   fprintf(out, " * Change for dynamics: GPAR = %.3f, GDPAR = %.3f *\n\n", gpar, gdpar);
   fprintf(out, "                  --------------------------------------------------------\n");
   fprintf(out, "                    Norm      Chem        Ener/N     <rho>     |Psi(0,0)|^2\n");
   fprintf(out, "                  --------------------------------------------------------\n");
   fflush(out);

   if (Nstp == 0) {
      g = par * g_2d;
      gd = par * gd0;
   } else {
      g = 0.;
      gd = 0.;
   }

   fftw_init_threads();
   fftw_plan_with_nthreads(nthreads);
   initpsi(psi);
   initpot();
   gencoef();
   initpotdd(*tmpxi, *tmpyi, *tmpxj, *tmpyj);

   plan_forward = fftw_plan_dft_r2c_2d(Nx, Ny, *psidd2, psidd2fft, FFT_FLAG);
   plan_backward = fftw_plan_dft_c2r_2d(Nx, Ny, psidd2fft, *psidd2, FFT_FLAG);


   calcnorm(&norm, psi, tmpxi, tmpyi);
   calcmuen(&mu, &en, psi, dpsi, psidd2, tmpxi, tmpyi, tmpxj, tmpyj);
   calcrms(rms, psi, tmpxi, tmpyi);
   psi2 = cabs(psi[Nx2][Ny2]);
   psi2 *= psi2;
   fprintf(out, "Initial : %15.4f %11.5F %11.5F %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
   fflush(out);
   if (initout != NULL) {
      sprintf(filename, "%s.txt", initout);
      file = fopen(filename, "w");
      outpsi2xy(psi, file);
      fclose(file);

      sprintf(filename, "%s1d_x.txt", initout);
      file = fopen(filename, "w");
      outdenx(psi, *tmpyi, file);
      fclose(file);

      sprintf(filename, "%s1d_y.txt", initout);
      file = fopen(filename, "w");
      outdeny(psi, *tmpxi, file);
      fclose(file);
   }
   if (rmsout != NULL) {
      fprintf(filerms, " Real time propagation 2dXY,   OPTION = %d\n\n", opt);
      fprintf(filerms, "                  --------------------------------------------------------\n");
      fprintf(filerms, "Values of rms size:        <x>             <y>             <rho>\n");
      fprintf(filerms, "                  --------------------------------------------------------\n");
      fprintf(filerms, "           Initial:%14.5F   %13.5F   %13.5F\n", rms[1], rms[2], rms[0]);
      fflush(filerms);
   }

   if (Nstp != 0) {
      double g_stp = par * g_2d / (double) Nstp;
      double gd_stp = par * gd0 / (double) Nstp;
      g = 0.;
      gd = 0.;
      gettimeofday(&iter_start, NULL);
      for (cnti = 0; cnti < Nstp; cnti ++) {
         g += g_stp;
         gd += gd_stp;
         calcnu(psi, psidd2);
         calclux(psi, cbeta);
         calcluy(psi, cbeta);
      }
      gettimeofday(&iter_stop, NULL);
      iter_time += (double) (((iter_stop.tv_sec - iter_start.tv_sec) * 1000 + (iter_stop.tv_usec - iter_start.tv_usec)/1000.0) + 0.5);

      calcnorm(&norm, psi, tmpxi, tmpyi);
      calcmuen(&mu, &en, psi, dpsi, psidd2, tmpxi, tmpyi, tmpxj, tmpyj);
      calcrms(rms, psi, tmpxi, tmpyi);
      psi2 = cabs(psi[Nx2][Ny2]) * cabs(psi[Nx2][Ny2]);
      fprintf(out, "After NSTP iter.:%8.4f %11.5F %11.5F %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
      fflush(out);
      if (Nstpout != NULL) {
         sprintf(filename, "%s.txt", Nstpout);
         file = fopen(filename, "w");
         outpsi2xy(psi, file);
         fclose(file);

         sprintf(filename, "%s1d_x.txt", Nstpout);
         file = fopen(filename, "w");
         outdenx(psi, *tmpyi, file);
         fclose(file);

         sprintf(filename, "%s1d_y.txt", Nstpout);
         file = fopen(filename, "w");
         outdeny(psi, *tmpxi, file);
         fclose(file);
      }
      if (rmsout != NULL) {
         fprintf(filerms, "  After NSTP iter.:%14.5F   %13.5F   %13.5F\n", rms[1], rms[2], rms[0]);
         fflush(filerms);
      }
   } else {
      g = par * g_2d;
      gd = par * gd0;
   }

   if (dynaout != NULL) {
      sprintf(filename, "%s.txt", dynaout);
      dyna = fopen(filename, "w");
   } else dyna = NULL;

   gettimeofday(&iter_start, NULL);
   for (cnti = 0; cnti < Npas; cnti ++) {
      calcnu(psi, psidd2);
      calclux(psi, cbeta);
      calcluy(psi, cbeta);
      if ((dynaout != NULL) && (cnti % outstpt == 0)) {
         calcrms(rms, psi, tmpxi, tmpyi);
         fprintf(dyna, "%8le   %8le   %8le   %8le\n", cnti * dt * par, rms[1], rms[2], rms[0]);
         fflush(dyna);
      }
   }
   gettimeofday(&iter_stop, NULL);
   iter_time += (double) (((iter_stop.tv_sec - iter_start.tv_sec) * 1000 + (iter_stop.tv_usec - iter_start.tv_usec)/1000.0) + 0.5);

   calcnorm(&norm, psi, tmpxi, tmpyi);
   calcmuen(&mu, &en, psi, dpsi, psidd2, tmpxi, tmpyi, tmpxj, tmpyj);
   calcrms(rms, psi, tmpxi, tmpyi);
   psi2 = cabs(psi[Nx2][Ny2]) * cabs(psi[Nx2][Ny2]);
   fprintf(out, "After NPAS iter.:%8.4f %11.5F %11.5F %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
   fflush(out);
   if (Npasout != NULL) {
      sprintf(filename, "%s.txt", Npasout);
      file = fopen(filename, "w");
      outpsi2xy(psi, file);
      fclose(file);

      sprintf(filename, "%s1d_x.txt", Npasout);
      file = fopen(filename, "w");
      outdenx(psi, *tmpyi, file);
      fclose(file);

      sprintf(filename, "%s1d_y.txt", Npasout);
      file = fopen(filename, "w");
      outdeny(psi, *tmpxi, file);
      fclose(file);
   }
   if (rmsout != NULL) {
      fprintf(filerms, "  After NPAS iter.:%14.5F   %13.5F   %13.5F\n", rms[1], rms[2], rms[0]);
      fflush(filerms);
   }

   g = gpar * g;
   gd = gdpar * gd;

   gettimeofday(&iter_start, NULL);
   for (cnti = 0; cnti < Nrun; cnti ++) {
      calcnu(psi, psidd2);
      calclux(psi, cbeta);
      calcluy(psi, cbeta);
      if ((dynaout != NULL) && (cnti % outstpt == 0)) {
         calcrms(rms, psi, tmpxi, tmpyi);
         fprintf(dyna, "%8le   %8le   %8le   %8le\n", (cnti * dt  + dt * Npas) * par, rms[1], rms[2], rms[0]);
      }
   }
   gettimeofday(&iter_stop, NULL);
   iter_time += (double) (((iter_stop.tv_sec - iter_start.tv_sec) * 1000 + (iter_stop.tv_usec - iter_start.tv_usec)/1000.0) + 0.5);

   if (dynaout != NULL) fclose(dyna);
   calcnorm(&norm, psi, tmpxi, tmpyi);
   calcmuen(&mu, &en, psi, dpsi, psidd2, tmpxi, tmpyi, tmpxj, tmpyj);
   calcrms(rms, psi, tmpxi, tmpyi);
   psi2 = cabs(psi[Nx2][Ny2]) * cabs(psi[Nx2][Ny2]);
   fprintf(out, "After NRUN iter.:%8.4f %11.5F %11.5F %10.5f %10.5f\n", norm, mu / par, en / par, *rms, psi2);
   fflush(out);
   if (Nrunout != NULL) {
      sprintf(filename, "%s.txt", Nrunout);
      file = fopen(filename, "w");
      outpsi2xy(psi, file);
      fclose(file);

      sprintf(filename, "%s1d_x.txt", Nrunout);
      file = fopen(filename, "w");
      outdenx(psi, *tmpyi, file);
      fclose(file);

      sprintf(filename, "%s1d_y.txt", Nrunout);
      file = fopen(filename, "w");
      outdeny(psi, *tmpxi, file);
      fclose(file);
   }
   if (rmsout != NULL) {
      fprintf(filerms, "  After NRUN iter.:%14.5F   %13.5F   %13.5F\n", rms[1], rms[2], rms[0]);
      fprintf(filerms, "                  --------------------------------------------------------\n");
   }
   if (rmsout != NULL) fclose(filerms);

   fprintf(out, "                  --------------------------------------------------------\n\n");

   free_double_vector(rms);

   free_double_vector(x);
   free_double_vector(y);

   free_double_vector(x2);
   free_double_vector(y2);

   free_double_matrix(pot);
   free_double_matrix(potdd);
   free_complex_matrix(psi);
   free_double_matrix(psidd2);

   free_double_matrix(dpsi);

   free_complex_vector(calphax);
   free_complex_vector(calphay);
   free_complex_matrix(cbeta);
   free_complex_vector(cgammax);
   free_complex_vector(cgammay);

   free_double_matrix(tmpxi);
   free_double_matrix(tmpyi);
   free_double_matrix(tmpxj);
   free_double_matrix(tmpyj);

   fftw_destroy_plan(plan_forward);
   fftw_destroy_plan(plan_backward);

   gettimeofday(&stop, NULL);
   wall_time = (double) (((stop.tv_sec - start.tv_sec) * 1000 + (stop.tv_usec - start.tv_usec)/1000.0) + 0.5);
   init_time = wall_time - iter_time;
   fprintf(out, " Initialization/allocation wall-clock time: %.3f seconds\n", init_time / 1000.);
   fprintf(out, " Calculation (iterations) wall-clock time: %.3f seconds\n", iter_time / 1000.);

   if(output != NULL) fclose(out);

   return (EXIT_SUCCESS);
}

/**
 *    Reading input parameters from the configuration file.
 */
void readpar(void) {
   char *cfg_tmp;

   if ((cfg_tmp = cfg_read("OPTION")) == NULL) {
      fprintf(stderr, "OPTION is not defined in the configuration file\n");
      exit(EXIT_FAILURE);
   }
   opt = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("NATOMS")) == NULL) {
      fprintf(stderr, "NATOMS is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Na = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("G0")) == NULL) {
      if ((cfg_tmp = cfg_read("AHO")) == NULL) {
         fprintf(stderr, "AHO is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      aho = atof(cfg_tmp);

      if ((cfg_tmp = cfg_read("AS")) == NULL) {
         fprintf(stderr, "AS is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      as = atof(cfg_tmp);

      g0 = 4. * pi * as * Na * BOHR_RADIUS / aho;
   } else {
      g0 = atof(cfg_tmp);
   }

   if ((cfg_tmp = cfg_read("GPAR")) == NULL) {
      fprintf(stderr, "GPAR is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   gpar = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("GDD0")) == NULL) {
      if ((cfg_tmp = cfg_read("AHO")) == NULL) {
         fprintf(stderr, "AHO is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      aho = atof(cfg_tmp);

      if ((cfg_tmp = cfg_read("ADD")) == NULL) {
         fprintf(stderr, "ADD is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      add = atof(cfg_tmp);

      gd0 = 3. * add * Na * BOHR_RADIUS / aho;
   } else {
      gd0 = atof(cfg_tmp);
   }

   if ((cfg_tmp = cfg_read("GDDPAR")) == NULL) {
      fprintf(stderr, "GDDPAR is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   gdpar = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("NX")) == NULL) {
      fprintf(stderr, "NX is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Nx = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("NY")) == NULL) {
      fprintf(stderr, "NY is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Ny = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("DX")) == NULL) {
      fprintf(stderr, "DX is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   dx = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("DY")) == NULL) {
      fprintf(stderr, "DY is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   dy = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("DT")) == NULL) {
      fprintf(stderr, "DT is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   dt = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("GAMMA")) == NULL) {
      fprintf(stderr, "GAMMA is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   vgamma = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("NU")) == NULL) {
      fprintf(stderr, "NU is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   vnu = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("D_Z")) == NULL) {
      fprintf(stderr, "D_Z is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   d_z = atof(cfg_tmp);

   if ((cfg_tmp = cfg_read("NSTP")) == NULL) {
      fprintf(stderr, "NSTP is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Nstp = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("NPAS")) == NULL) {
      fprintf(stderr, "NPAS is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Npas = atol(cfg_tmp);

   if ((cfg_tmp = cfg_read("NRUN")) == NULL) {
      fprintf(stderr, "NRUN is not defined in the configuration file.\n");
      exit(EXIT_FAILURE);
   }
   Nrun = atol(cfg_tmp);

   output = cfg_read("OUTPUT");
   rmsout = cfg_read("RMSOUT");
   dynaout = cfg_read("DYNAOUT");
   initout = cfg_read("INITOUT");
   Nstpout = cfg_read("NSTPOUT");
   Npasout = cfg_read("NPASOUT");
   Nrunout = cfg_read("NRUNOUT");

   if ((initout != NULL) || (Nstpout != NULL) || (Npasout != NULL) || (Nrunout != NULL)) {
      if ((cfg_tmp = cfg_read("OUTSTPX")) == NULL) {
         fprintf(stderr, "OUTSTPX is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      outstpx = atol(cfg_tmp);

      if ((cfg_tmp = cfg_read("OUTSTPY")) == NULL) {
         fprintf(stderr, "OUTSTPY is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      outstpy = atol(cfg_tmp);
   }

   if (dynaout != NULL) {
      if ((cfg_tmp = cfg_read("OUTSTPT")) == NULL) {
         fprintf(stderr, "OUTSTPT is not defined in the configuration file.\n");
         exit(EXIT_FAILURE);
      }
      outstpt = atol(cfg_tmp);
   }

   return;
}

/**
 *    Initialization of the space mesh and the initial wave function.
 *    psi - array with the wave function values
 */
void initpsi(double complex **psi) {
   long cnti, cntj;
   double cpsi;
   double tmp;
   FILE *file;

   cpsi = sqrt(pi * sqrt(1. / (vgamma * vnu)));

   #pragma omp parallel for private(cnti)
   for (cnti = 0; cnti < Nx; cnti ++) {
      x[cnti] = (cnti - Nx2) * dx;
      x2[cnti] = x[cnti] * x[cnti];
   }

   #pragma omp parallel for private(cntj)
   for (cntj = 0; cntj < Ny; cntj ++) {
      y[cntj] = (cntj - Ny2) * dy;
      y2[cntj] = y[cntj] * y[cntj];
   }

   if (Nstp != 0) {
      #pragma omp parallel for private(cnti, cntj, tmp)
      for (cnti = 0; cnti < Nx; cnti ++) {
         for (cntj = 0; cntj < Ny; cntj ++) {
            tmp = exp(- 0.5 * (vgamma * x2[cnti] + vnu * y2[cntj]));
            psi[cnti][cntj] = tmp / cpsi;
         }
      }
   }

   if (Nstp == 0) {
      if ((file = fopen("imag2dXY-den.txt", "r")) == NULL) {    /* open a text file for reading */
         printf("Run the program using the input file to read. i.g.: imag2dXY-den.txt\n");
         exit(1);               /*couldn't open the requested file!*/
      }
      file = fopen("imag2dXY-den.txt", "r");
      for (cnti = 0; cnti < Nx; cnti ++) {
         for (cntj = 0; cntj < Ny; cntj ++) {
            if(fscanf(file,"%lf %lf %lf\n", &x[cnti], &y[cntj], &tmp)) {
               psi[cnti][cntj] = sqrt(tmp);
            }
         }
      }
      fclose(file);
   }

   return;
}

/**
 *    Initialization of the potential.
 */
void initpot() {
   long cnti, cntj;
   double vnu2, vgamma2;

   vnu2 = vnu * vnu;
   vgamma2 = vgamma * vgamma;

   #pragma omp parallel for private(cnti, cntj)
   for (cnti = 0; cnti < Nx; cnti ++) {
      for (cntj = 0; cntj < Ny; cntj ++) {
         pot[cnti][cntj] = 0.5 * par * (vgamma2 * x2[cnti] + vnu2 * y2[cntj]);
      }
   }

   return;
}

/**
 *    Crank-Nicolson scheme coefficients generation.
 */
void gencoef(void) {
   long cnti;

   Ax0 = 1. + I * dt / dx2 / (3. - par);
   Ay0 = 1. + I * dt / dy2 / (3. - par);

   Ax0r = 1. - I * dt / dx2 / (3. - par);
   Ay0r = 1. - I * dt / dy2 / (3. - par);

   Ax = - 0.5 * I * dt / dx2 / (3. - par);
   Ay = - 0.5 * I * dt / dy2 / (3. - par);

   calphax[Nx - 2] = 0.;
   cgammax[Nx - 2] = - 1. / Ax0;
   for (cnti = Nx - 2; cnti > 0; cnti --) {
      calphax[cnti - 1] = Ax * cgammax[cnti];
      cgammax[cnti - 1] = - 1. / (Ax0 + Ax * calphax[cnti - 1]);
   }

   calphay[Ny - 2] = 0.;
   cgammay[Ny - 2] = - 1. / Ay0;
   for (cnti = Ny - 2; cnti > 0; cnti --) {
      calphay[cnti - 1] = Ay * cgammay[cnti];
      cgammay[cnti - 1] = - 1. / (Ay0 + Ay * calphay[cnti - 1]);
   }

   return;
}

/**
 *    Initialization of the dipolar potential.
 *    kx  - array with the space mesh values in the x-direction in the K-space
 *    ky  - array with the space mesh values in the y-direction in the K-space
 *    kx2 - array with the squared space mesh values in the x-direction in the
 *          K-space
 *    ky2 - array with the squared space mesh values in the y-direction in the
 *          K-space
 */
void initpotdd(double *kx, double *ky, double *kx2, double *ky2) {
   long cnti, cntj;
   double dkx, dky;
   double xi, xi2;

   dkx = 2. * pi / (Nx * dx);
   dky = 2. * pi / (Ny * dy);

   for (cnti = 0; cnti < Nx2; cnti ++) kx[cnti] = cnti * dkx;
   for (cnti = 0; cnti < Nx2; cnti ++) kx[cnti + Nx2] = (cnti - Nx2) * dkx;
   for (cntj = 0; cntj < Ny2; cntj ++) ky[cntj] = cntj * dky;
   for (cntj = 0; cntj < Ny2; cntj ++) ky[cntj + Ny2] = (cntj - Ny2) * dky;

   for (cnti = 0; cnti < Nx; cnti ++) kx2[cnti] = kx[cnti] * kx[cnti];
   for (cntj = 0; cntj < Ny; cntj ++) ky2[cntj] = ky[cntj] * ky[cntj];

   #pragma omp parallel for private(cnti, cntj, xi, xi2)
   for (cnti = 0; cnti < Nx; cnti ++) {
      for (cntj = 0; cntj < Ny; cntj ++) {
         xi2 = (kx2[cnti] + ky2[cntj]) * d_z * d_z / 2.;
         xi = sqrt(xi2);
         if (xi >= XI_LIMIT) {
            potdd[cnti][cntj] = POTDD_ASYMPTOTIC(xi);
         } else {
            potdd[cnti][cntj] = 2. - 3. * sqrt(pi) * xi * exp(xi2) * erfc(xi);
         }
         potdd[cnti][cntj] *= (4. * pi / 3.) / (sqrt(2. * pi) * d_z);
      }
   }

   return;
}

/**
 *    Calculation of the wave function norm and normalization.
 *    norm - wave function norm
 *    psi  - array with the wave function values
 *    tmpx - temporary array
 *    tmpy - temporary array
 */
void calcnorm(double *norm, double complex **psi, double **tmpx, double **tmpy) {
   int threadid;
   long cnti, cntj;
   double tmp;

   #pragma omp parallel private(threadid, cnti, cntj, tmp)
   {
      threadid = omp_get_thread_num();

      #pragma omp for
      for (cnti = 0; cnti < Nx; cnti ++) {
         for (cntj = 0; cntj < Ny; cntj ++) {
            tmp = cabs(psi[cnti][cntj]);
            tmpy[threadid][cntj] = tmp * tmp;
         }
         (*tmpx)[cnti] = simpint(dy, tmpy[threadid], Ny);
      }
   }

   *norm = sqrt(simpint(dx, *tmpx, Nx));
   // tmp = 1. / *norm;

   // for (cnti = 0; cnti < Nx; cnti ++) {
   //    for (cntj = 0; cntj < Ny; cntj ++) {
   //       psi[cnti][cntj] *= tmp;
   //    }
   // }

   return;
}

/**
 *    Calculation of the chemical potential and energy.
 *    mu     - chemical potential
 *    en     - energy
 *    psi    - array with the wave function values
 *    dpsi   - temporary array
 *    psidd2 - array with the squared wave function values
 *    tmpxi  - temporary array
 *    tmpyi  - temporary array
 *    tmpxj  - temporary array
 *    tmpyj  - temporary array
 */
void calcmuen(double *mu, double *en, double complex **psi, double **dpsi, double **psidd2, double **tmpxi, double **tmpyi, double **tmpxj, double **tmpyj) {
   int threadid;
   long cnti, cntj;
   double psi2, psi2lin, psidd2lin, dpsi2;

   calcpsidd2(psi, psidd2);

   #pragma omp parallel private(threadid, cnti, cntj, psi2, psi2lin, psidd2lin, dpsi2)
   {
      threadid = omp_get_thread_num();

      #pragma omp for
      for (cntj = 0; cntj < Ny; cntj ++) {
         for (cnti = 0; cnti < Nx; cnti ++) {
            tmpxi[threadid][cnti] = cabs(psi[cnti][cntj]);
         }
         diff(dx, tmpxi[threadid], tmpxj[threadid], Nx);
         for (cnti = 0; cnti < Nx; cnti ++) {
            dpsi[cnti][cntj] = tmpxj[threadid][cnti] * tmpxj[threadid][cnti];
         }
      }

      #pragma omp for
      for (cnti = 0; cnti < Nx; cnti ++) {
         for (cntj = 0; cntj < Ny; cntj ++) {
            tmpyi[threadid][cntj] = cabs(psi[cnti][cntj]);
         }
         diff(dy, tmpyi[threadid], tmpyj[threadid], Ny);
         for (cntj = 0; cntj < Ny; cntj ++) {
            dpsi[cnti][cntj] += tmpyj[threadid][cntj] * tmpyj[threadid][cntj];
         }
      }

      #pragma omp barrier

      #pragma omp for
      for (cnti = 0; cnti < Nx; cnti ++) {
         for (cntj = 0; cntj < Ny; cntj ++) {
            psi2 = cabs(psi[cnti][cntj]);
            psi2 *= psi2;
            psi2lin = psi2 * g;
            psidd2lin = psidd2[cnti][cntj] * gd;
            dpsi2 = dpsi[cnti][cntj] / (3. - par);
            tmpyi[threadid][cntj] = (pot[cnti][cntj] + psi2lin + psidd2lin) * psi2 + dpsi2;
            tmpyj[threadid][cntj] = (pot[cnti][cntj] + 0.5 * psi2lin + 0.5 * psidd2lin) * psi2 + dpsi2;
         }
         (*tmpxi)[cnti] = simpint(dy, tmpyi[threadid], Ny);
         (*tmpxj)[cnti] = simpint(dy, tmpyj[threadid], Ny);
      }
   }

   *mu = simpint(dx, *tmpxi, Nx);
   *en = simpint(dx, *tmpxj, Nx);

   return;
}

/**
 *    Calculation of squared wave function values for dipole-dipole
 *    interaction.
 *    psi    - array with the wave function values
 *    psidd2 - array with the squared wave function values
 */
void calcpsidd2(double complex **psi, double **psidd2) {
   long cnti, cntj;
   double tmp;
   fftw_complex *psidd2fft = (fftw_complex *) *psidd2;

   #pragma omp parallel for private(cnti, cntj, tmp)
   for (cnti = 0; cnti < Nx; cnti ++) {
      for (cntj = 0; cntj < Ny; cntj ++) {
         tmp = cabs(psi[cnti][cntj]);
         psidd2[cnti][cntj] = tmp * tmp;
      }
   }

   fftw_execute_dft_r2c(plan_forward, *psidd2, psidd2fft);

   #pragma omp parallel for private(cnti, cntj)
   for (cnti = 0; cnti < Nx; cnti ++) {
      for (cntj = 0; cntj < (Ny / 2) + 1; cntj ++) {
         psidd2fft[cnti * (Ny2 + 1) + cntj][0] *= potdd[cnti][cntj];
         psidd2fft[cnti * (Ny2 + 1) + cntj][1] *= potdd[cnti][cntj];
      }
   }

   fftw_execute_dft_c2r(plan_backward, psidd2fft, *psidd2);

   #pragma omp parallel for private(cnti, cntj)
   for (cnti = 0; cnti < Nx; cnti ++) {
      for (cntj = 0; cntj < Ny; cntj ++) {
         psidd2[cnti][cntj] /= Nx * Ny;
      }
   }

   #pragma omp parallel for private(cntj)
   for (cntj = 0; cntj < Ny - 1; cntj ++) {
      psidd2[Nx - 1][cntj] = psidd2[0][cntj];
   }

   #pragma omp parallel for private(cnti)
   for (cnti = 0; cnti < Nx - 1; cnti ++) {
      psidd2[cnti][Ny - 1] = psidd2[cnti][0];
   }

   return;
}

/**
 *    Calculation of the root mean square radius.
 *    rms  - root mean square radius
 *    psi  - array with the wave function values
 *    tmpx - temporary array
 *    tmpy - temporary array
 */
void calcrms(double *rms, double complex **psi, double **tmpx, double **tmpy) {
   int threadid;
   long cnti, cntj;
   double psi2;

   #pragma omp parallel private(threadid, cnti, cntj, psi2)
   {
      threadid = omp_get_thread_num();

      #pragma omp for
      for (cntj = 0; cntj < Ny; cntj ++) {
         for (cnti = 0; cnti < Nx; cnti ++) {
            psi2 = cabs(psi[cnti][cntj]);
            psi2 *= psi2;
            tmpx[threadid][cnti] = x2[cnti] * psi2;
         }
         (*tmpy)[cntj] = simpint(dx, tmpx[threadid], Nx);
      }
      #pragma omp barrier

      #pragma omp single
      rms[1] = sqrt(simpint(dy, *tmpy, Ny));

      #pragma omp for
      for (cnti = 0; cnti < Nx; cnti ++) {
         for (cntj = 0; cntj < Ny; cntj ++) {
            psi2 = cabs(psi[cnti][cntj]);
            psi2 *= psi2;
            tmpy[threadid][cntj] = y2[cntj] * psi2;
         }
         (*tmpx)[cnti] = simpint(dy, tmpy[threadid], Ny);
      }
      #pragma omp barrier

      #pragma omp single
      rms[2] = sqrt(simpint(dx, *tmpx, Nx));
   }

   rms[0] = sqrt(rms[1] * rms[1] + rms[2] * rms[2]);

   return;
}

/**
 *    Time propagation with respect to H1 (part of the Hamiltonian without
 *    spatial derivatives).
 *    psi    - array with the wave function values
 *    psidd2 - array with the squared wave function values
 */
void calcnu(double complex **psi, double **psidd2) {
   long cnti, cntj;
   double psi2, psi2lin, psidd2lin, tmp;

   calcpsidd2(psi, psidd2);

   #pragma omp parallel for private(cnti, cntj, psi2, psi2lin, psidd2lin, tmp)
   for (cnti = 0; cnti < Nx; cnti ++) {
      for (cntj = 0; cntj < Ny; cntj ++) {
         psi2 = cabs(psi[cnti][cntj]);
         psi2 *= psi2;
         psi2lin = psi2 * g;
         psidd2lin = psidd2[cnti][cntj] * gd;
         tmp = dt * (pot[cnti][cntj] + psi2lin + psidd2lin);
         psi[cnti][cntj] *= cexp(- I * tmp);
      }
   }

   return;
}

/**
 *    Time propagation with respect to H2 (x-part of the Laplacian).
 *    psi   - array with the wave function values
 *    cbeta - Crank-Nicolson scheme coefficients
 */
void calclux(double complex **psi, double complex **cbeta) {
   int threadid;
   long cnti, cntj;
   double complex c;

   #pragma omp parallel private(threadid, cnti, cntj, c)
   {
      threadid = omp_get_thread_num();

      #pragma omp for
      for (cntj = 0; cntj < Ny; cntj ++) {
         cbeta[threadid][Nx - 2] = psi[Nx - 1][cntj];
         for (cnti = Nx - 2; cnti > 0; cnti --) {
            c = - Ax * psi[cnti + 1][cntj] + Ax0r * psi[cnti][cntj] - Ax * psi[cnti - 1][cntj];
            cbeta[threadid][cnti - 1] =  cgammax[cnti] * (Ax * cbeta[threadid][cnti] - c);
         }
         psi[0][cntj] = 0.;
         for (cnti = 0; cnti < Nx - 2; cnti ++) {
            psi[cnti + 1][cntj] = calphax[cnti] * psi[cnti][cntj] + cbeta[threadid][cnti];
         }
         psi[Nx - 1][cntj] = 0.;
      }
   }

   return;
}


/**
 *    Time propagation with respect to H3 (y-part of the Laplacian).
 *    psi   - array with the wave function values
 *    cbeta - Crank-Nicolson scheme coefficients
 */
void calcluy(double complex **psi, double complex **cbeta) {
   int threadid;
   long cnti, cntj;
   double complex c;

   #pragma omp parallel private(threadid, cnti, cntj, c)
   {
      threadid = omp_get_thread_num();

      #pragma omp for
      for (cnti = 0; cnti < Nx; cnti ++) {
         cbeta[threadid][Ny - 2] = psi[cnti][Ny - 1];
         for (cntj = Ny - 2; cntj > 0; cntj --) {
            c = - Ay * psi[cnti][cntj + 1] + Ay0r * psi[cnti][cntj] - Ay * psi[cnti][cntj - 1];
            cbeta[threadid][cntj - 1] =  cgammay[cntj] * (Ay * cbeta[threadid][cntj] - c);
         }
         psi[cnti][0] = 0.;
         for (cntj = 0; cntj < Ny - 2; cntj ++) {
            psi[cnti][cntj + 1] = calphay[cntj] * psi[cnti][cntj] + cbeta[threadid][cntj];
         }
         psi[cnti][Ny - 1] = 0.;
      }
   }

   return;
}

void outpsi2xy(double complex **psi, FILE *file) {
   long cnti, cntj;

   for(cnti = 0; cnti < Nx; cnti += outstpx) {
      for(cntj = 0; cntj < Ny; cntj += outstpy) {
         fprintf(file, "%8le %8le %8le\n", x[cnti], y[cntj], cabs(psi[cnti][cntj]) * cabs(psi[cnti][cntj]));
      }
      fprintf(file, "\n");
      fflush(file);
   }

   return;
}

void outdenx(double complex **psi, double *tmp, FILE *file) {
   long cnti, cntj;

   for (cnti = 0; cnti < Nx; cnti += outstpx) {
      for (cntj = 0; cntj < Ny; cntj ++) {
         tmp[cntj] = cabs(psi[cnti][cntj]) * cabs(psi[cnti][cntj]);
      }
      fprintf(file, "%8le %8le\n", x[cnti], simpint(dy, tmp, Ny));
      fflush(file);
   }
}

void outdeny(double complex **psi, double *tmp, FILE *file) {
   long cnti, cntj;

   for (cntj = 0; cntj < Ny; cntj += outstpy) {
      for (cnti = 0; cnti < Nx; cnti ++) {
         tmp[cnti] = cabs(psi[cnti][cntj]) * cabs(psi[cnti][cntj]);
      }
      fprintf(file, "%8le %8le\n", y[cntj], simpint(dx, tmp, Nx));
      fflush(file);
   }
}
